# coding: utf-8
"""
    InputDataUtils.py

    Migration tool aimed to be compatible with SerialDataUtils.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import os
import re
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting, set_path_length
from SerialDataUtils import get_uv_filename, get_mtd_filename
from InputData import InputData

teller = None
current_data = None

def initialize(debug=False):
    global teller
    if teller is None:
        from InputProcess.InputTeller import InputTeller
        this_dir = os.path.dirname( os.path.abspath( __file__ ) )
        log_folder = os.path.abspath( this_dir + '/../../log' )
        teller = InputTeller(log_folder=log_folder, debug=debug)

def terminate():
    # without this, this parent process won't stop.
    global current_data
    global teller
    del current_data
    del teller

def _get_data(in_folder, debug=True):
    global current_data
    initialize(debug=debug)

    if current_data is None:
        current_data = InputData(in_folder, teller, debug=debug)
        current_data.prepare()

    return current_data

def clear_current_data():
    global current_data
    current_data = None

def update_path_length(comments):
    measurement_date = None
    if comments is not None:
        for line in comments:
            if line.find('Date') > 0:
                date_re = re.compile(r'(\d+/\d+/\d+)')
                m = date_re.search(line)
                if m:
                    measurement_date = int(m.group(1).replace('/', ''))
                    break
    set_setting('measurement_date', measurement_date)
    set_path_length(measurement_date)

def load_intensity_files( in_folder, logger=None, debug=True ):
    data = _get_data(in_folder)
    data_array = data.make_redundant_xray_data()
    update_path_length(data.comments)
    return data_array, data.files

def load_uv_array( conc_folder, conc_file=None, column_header=False ):
    disable_uv_data = get_setting('disable_uv_data')

    if disable_uv_data:
        uv_data = None
    else:
        data = _get_data(conc_folder)
        uv_data = data.uv_data

    if uv_data is None:
        data_array = None
        lvector = None
        conc_file = None
        col_header = None
    else:
        data_array = uv_data.data
        lvector = uv_data.vector
        folder_info = data.folder_info
        conc_file = folder_info.conc_file
        col_header = [None]*(data_array.shape[1]+1)     # dummies; assuming that col_header is not used

    if column_header:
        return data_array, lvector, conc_file, col_header
    else:
        return data_array, lvector, conc_file



"""
    Models/Stochastic/CfOptimizationStudy.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.DataStructure.SvdDenoise import get_denoised_data
from molass_legacy.Models.ModelUtils import compute_cy_list, compute_area_props
from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier
from SecTheory.SecPDF import BidirectFft

def cf_optimization_study_impl(advanced_frame, devel=True):
    if devel:
        from importlib import reload
        import Models.Stochastic.LognormalPoreColumn as LognormalPoreColumn
        reload(LognormalPoreColumn)
    from molass_legacy.Models.Stochastic.LognormalPoreColumn import LognormalPoreColumn
    editor = advanced_frame.editor
    editor_frame = editor.get_current_frame()
    model = editor_frame.model
    print("estimate_psd_impl", model.get_name(), model.__class__)
    params_array = editor.get_current_params_array()

    fx = editor_frame.fx
    x = editor_frame.x
    y = editor_frame.y
    uv_y = editor_frame.uv_y

    D, E, qv, ecurve = editor.sd.get_xr_data_separate_ly()
    peak_region = ecurve.get_peak_region(sigma_scale=5)
    print("peak_region=", peak_region)

    def plot_cy_list(cy_list, title, test_y=None):
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title(title)
            ax.plot(x, y)
            for cy in cy_list:
                ax.plot(x, cy, ":", label="component-%d" % (len(cy_list)))
            ty = np.sum(cy_list, axis=0)
            ax.plot(x, ty, ":", color="red", label="model total")
            for x_ in peak_region:
                ax.axvline(x=x_+x[0], color="green")
            if test_y is not None:
                ax.plot(x, test_y, ":", color="cyan", label="test curve")
            fig.tight_layout()
            plt.show()

    slice_ = slice(*[int(x_) for x_ in peak_region])
    num_components = len(params_array)
    print("num_components=", num_components)
    M_ = get_denoised_data(D[:,slice_], rank=num_components)
    E_ = E[:,slice_]
    cy_list = compute_cy_list(model, fx, params_array)
    plot_cy_list(cy_list, "initial state")

    min_msd = None
    column = LognormalPoreColumn()
    init_scales = None
    last_rgs = None
    for k in range(5):
        C = np.array(cy_list)
        C_ = C[:,slice_]
        Cinv = np.linalg.pinv(C_)
        P_ = M_ @ Cinv
        Minv = np.linalg.pinv(M_)
        W = Minv @ P_
        Pe = np.sqrt(E_**2 @ W**2)

        rg_list = []
        for j, p_ in enumerate(P_.T):
            data = np.array([qv, p_, Pe[:,j]]).T
            sg = SimpleGuinier(data)
            rg_list.append(sg.Rg)

        x0 = 300
        scale = 3
        print("rg_list=", rg_list)
        cy_list = []
        for rg in rg_list:
            cy = column.compute_curve(fx, x0, scale, rg)
            cy_list.append(cy)

        bifft = BidirectFft(fx)
        z = bifft.compute_z(y)
        w = bifft.get_w()
        u = bifft.compute_y(z)

        with plt.Dp():
            fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18, 5))
            ax1.set_title("PDF Domain")
            ax1.plot(x, y)
            for cy in cy_list:
                ax1.plot(x, cy, ":")
            for x_ in peak_region:
                ax1.axvline(x=x_+x[0], color="green")
            ax2.set_title("CF Real Domain")
            ax2.plot(w, np.real(z))
            ax3.set_title("CF Imaginary Domain")
            ax3.plot(w, np.imag(z))
            ax1.plot(x, u, ":", color="cyan")
            fig.tight_layout()
            plt.show()
        break


"""
    Distance.NormalizedRmsd.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import numpy as np

def normalized_rmsd(y1, y2, weights=None, adjust=0, debug=False):
    diff = y1 - y2
    if weights is not None:
        diff *= weights
    return np.log10(np.sqrt(np.mean(diff**2))) + adjust


# coding: utf-8
"""
    ExtrapolationAnimation.py

    Copyright (c) 2018-2019, SAXS Team, KEK-PF
"""
import os
import numpy as np
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import set_icon
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import matplotlib.animation as animation
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurMatplotlib import get_color
from DevSettings import get_dev_setting

TOO_SMALL_TO_PLOT   = 1e-12

def format_func(value, tick_number):
    return '$10^{%d}$' % (value)

class ExtrapolationAnimationDialog(Dialog):
    def __init__( self, parent, row, q, ridge, anim_data, penalty_weights, ignore_bq=0 ):
        self.grab = 'local'     # used in grab_set
        self.q  = q
        self.ridge = ridge
        self.parent = parent
        self.row = row
        self.anim_data1 = anim_data[0]
        self.anim_data2 = anim_data[1]
        self.min_data = np.log10( max( TOO_SMALL_TO_PLOT, np.min(self.anim_data2) ) )
        self.max_data = np.log10( np.max(self.anim_data2) )
        print('min_data=', self.min_data, 'max_data=', self.max_data)
        self.row_penalty_weights = [1.0] + penalty_weights
        self.ignore_bq = ignore_bq
        self.add_conc_const = get_dev_setting('add_conc_const')
        self.aq_smoothness = parent.aq_smoothness
        self.aq_positivity = parent.aq_positivity
        self.pause = False

    def show(self):
        title = "Extrapolation Animation"
        Dialog.__init__( self, self.parent, title )

    def body(self, body_frame):
        set_icon( self )

        cframe = Tk.Frame(self)
        cframe.pack()

        fig = plt.figure( figsize=(21,7) )
        gs = GridSpec(1, 3)
        ax1 = fig.add_subplot( gs[0,0] )
        self.ax2 = ax2 = fig.add_subplot( gs[0,1:] )
        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )
        self.mpl_canvas.mpl_connect('button_press_event', self.on_click)

        ax1.set_title("Changes of objective term values", fontsize=20)
        ax1.set_xlabel("Function call count")
        ax1.set_ylabel("Objective Values")
        if self.ignore_bq:
            curve_text = "curve: A(q)"
        else:
            curve_text = "curves: A(q), B(q)"
        ax2.set_title("Optimized scattering essense " + curve_text, fontsize=20)
        ax2.set_xlabel("Q")
        ax2.set_ylabel("$log_{10}(Intensity)$")
        ax1.yaxis.set_major_formatter(plt.FuncFormatter(format_func))
        ax2.yaxis.set_major_formatter(plt.FuncFormatter(format_func))

        # draw ax2 first to get ymin, ymax
        if not self.ignore_bq:
            ax2t = ax2.twinx()
            ax2t.set_ylabel("Intensity")
            ax2t.grid(False)
        r_color = get_color(0)
        aq_color = get_color(1)
        bq_color = 'pink'
        # note that anim_rec[0] > 0
        self.n  = np.array([ anim_rec1[0] for anim_rec1 in self.anim_data1 ])
        anim_rec1 = self.anim_data1[0]
        P  = anim_rec1[1]
        ax2.plot( self.q, np.log10(self.ridge), color=get_color(0))
        aq_y = np.log10(P[:,0])
        aq_curve, = ax2.plot( self.q, aq_y, color=aq_color, label='A(q)', alpha=1 )

        if not self.ignore_bq:
            bq_y = P[:,1]
            bq_curve, = ax2t.plot( self.q, bq_y, color=bq_color, label='B(q)' )

        # ax1.set_yscale("log", nonposy='clip')
        numerator = 0
        denominator = self.anim_data1[-1][0]
        xmin2, xmax2 = ax2.get_xlim()
        ymin2, ymax2 = ax2.get_ylim()
        tx = xmin2*0.4 + xmax2*0.6
        ty = ymin2*0.25 + ymax2*0.75
        text = ax2.text( tx, ty, r"$\frac{%d}{%d}$" % (numerator, denominator), fontsize=100, alpha=0.2 )
        tx = xmin2*0.95 + xmax2*0.05
        ty = ymin2*0.95 + ymax2*0.05
        self.guide_text = ax2.text( tx, ty, "", fontsize=50, alpha=0.1 )
        self.update_guide_text()

        ax2.legend(bbox_to_anchor=(1.0, 1.0), loc='upper right', fontsize=16)
        if not self.ignore_bq:
            ax2t.legend(bbox_to_anchor=(1.0, 0.92), loc='upper right', fontsize=16)

        self.chisqr_list = []
        anim_rec2 = self.anim_data2[0]
        self.chisqr_list.append(anim_rec2)
        c_data = np.log10( np.array(self.chisqr_list) )
        c_lines = []
        label_texts = [ 'Re-construction', 'A(q)-positivity', 'Base-drift', 'A(q)-smoothness', 'B(q)-smoothness' ]
        self.c_line_indeces = []
        for k in range(c_data.shape[1]):
            if k == 1:
                if not self.aq_positivity:
                    continue
            elif k == 2:
                if not self.add_conc_const:
                    continue
            elif k > 2:
                if not self.aq_smoothness:
                    continue
            scale = self.row_penalty_weights[k]
            # note that c_data already insludes scale
            c_line, = ax1.plot( [0], np.log10(c_data[:,k]), color=get_color(k), label=('%4.2f * ' % scale) + label_texts[k] )
            c_lines.append(c_line)
            self.c_line_indeces.append(k)

        _, ymax1 = ax1.get_ylim()
        ax1.set_xlim( 1, self.n[-1] )
        ymin1 = self.min_data - 0.5 if np.isfinite(self.min_data) else ymin2 - 2
        ymax1 = self.max_data + 2 if np.isfinite(self.max_data) else max(ymax2, ymax1) + 2

        ax1.set_ylim(ymin1, ymax1)
        ax1.set_xscale("log", nonposx='clip')
        ax1.legend(fontsize=16)

        fig.tight_layout()

        def init():
            anim_rec2 = self.anim_data2[0]
            self.chisqr_list = []
            self.chisqr_list.append(anim_rec2)
            text.set_text( r"$\frac{\ }{%d}$" % ( denominator) )
            aq_curve.set_xdata(np.ma.array( self.q, mask=True))
            aq_curve.set_ydata(np.ma.array( aq_y, mask=True))
            if not self.ignore_bq:
                bq_curve.set_xdata(np.ma.array( self.q, mask=True))
                bq_curve.set_ydata(np.ma.array( bq_y, mask=True))
            chisqr_array = np.array(self.chisqr_list)
            for k, c_line in enumerate(c_lines):
                c_line.set_xdata(np.ma.array( self.n[0:chisqr_array.shape[0]], mask=True) )
                c_line.set_ydata(np.ma.array( chisqr_array[:,k], mask=True))

            if self.ignore_bq:
                return tuple( [ text, self.guide_text, aq_curve ] + c_lines )
            else:
                return tuple( [ text, self.guide_text, aq_curve, bq_curve ] + c_lines )

        def animate( anim_tuple ):
            anim_rec1, anim_rec2 = anim_tuple
            numerator = anim_rec1[0]
            P = anim_rec1[1]

            text.set_text( r"$\frac{%d}{%d}$" % (numerator, denominator) )
            aq_curve.set_xdata( self.q )
            aq_curve.set_ydata( np.log10(P[:,0]) )
            # set 0 alpha for the first few iterations to avoid them to remain visible (matplotlib bug?)
            aq_curve.set_alpha( 0 if numerator < 3 else 1 )
            if not self.ignore_bq:
                bq_curve.set_xdata( self.q )
                bq_curve.set_ydata( P[:,1] )
            chisqr_array = np.array(self.chisqr_list)
            chisqr_array[chisqr_array < TOO_SMALL_TO_PLOT] = TOO_SMALL_TO_PLOT
            for k, c_line in enumerate(c_lines):
                k_ = self.c_line_indeces[k]
                c_line.set_xdata( self.n[0:chisqr_array.shape[0]] )
                c_line.set_ydata( np.log10(chisqr_array[:,k_]) )

            if self.ignore_bq:
                return tuple( [ text, self.guide_text, aq_curve ] + c_lines )
            else:
                return tuple( [ text, self.guide_text, aq_curve, bq_curve ] + c_lines )

        def index_generator():
            i = 0
            data_size = len(self.anim_data2)
            while i < data_size + 10:
                i_ = min(data_size-1, i)
                anim_rec2 = self.anim_data2[i_]
                if self.pause:
                    pass
                else:
                    if i == 0:
                        # already appended
                        pass
                    else:
                        if i_ == i:
                            self.chisqr_list.append(anim_rec2)
                    i += 1
                yield self.anim_data1[i_], anim_rec2

        self.anim = animation.FuncAnimation(fig, animate, index_generator, init_func=init,
                    interval=100, blit=True)

        # self.mpl_canvas.draw()

    def on_click(self, event):
        if event.inaxes is None:
            return
        self.pause ^= True
        self.update_guide_text()

    def update_guide_text(self):
        self.guide_text.set_text( "Click to restart" if self.pause else "Click to pause" )
        # note that self.guide_text must be included in the animate() func return
        # for it to be shown updated



# coding: utf-8
"""
    Guinier.py

    Copyright (c) 2016, Masatsuyo Takahashi, KEK-PF
"""
import sys
import numpy                as np
from bisect                 import bisect_right
from scipy.interpolate      import UnivariateSpline
from molass_legacy.KekLib.ExceptionTracebacker   import ExceptionTracebacker
from WLS_Rg                 import WLS_Rg, WLS_HeadTrend
from Quality                import Quality
from Result                 import Result
from SmoothSpline           import SmoothSpline
from BiComponentSolver      import BiComponentSolver, REQUIRED_BASIC_QUALITY

ANALYZABLE_STDEV_RATIO      = 0.04
IPI_MAX_STDEV_RATIO         = 0.01
# IPI_REQ_BASIC_QUALITY       = 0.9
IPI_CURVATURE_BOUNDARY      = -0.3      # -2.08 for ts008_ave_sub.dat
FIT_CONSISTENCY_PURE_BOUNDARY   = 0.9
AIC_SCORE_BOUNDARY          = 9.0
REQUIRED_QUALITY            = 0.2
HEAD_TREND_POLY_BOUNDARY    = -0.5
POLY_QRG_SCORE_BOUNDARY     = 1e-5
# POLY_QRG_SCORE_BOUNDARY     = 0.6     # 0.51 for ts008_ave_sub.dat
POLY_REQ_BASIC_CONDITION    = 0.9       # for Mei_subtRNA_00189_sub.dat, etc.
POLY_STRICT_STDEV_RATIO     = 0.007     # 0.009 for 
POLY_NARROW_RATIO_BOUNDARY  = 0.05      # 0.07 for AIMdim01_00226_sub.dat
POLY_WIDE_RATIO_BOUNDARY    = 0.8       # 0.700 for ts008_ave_sub.dat
POLY_ASSERTABLE_PROPORTION  = 0.92
POLY_ASSERTABLE_DIFF_RATIO  = 0.08      # Rg_diff= 0.137645266692 for OA_1.0_ave_sub.dat

DEBUG = False

class Guinier:
    def __init__( self, x, y, e, positive_ratio=None ):
        self.x      = x
        self.y      = y
        self.e      = e
        self.wls        = None
        self.interval   = None
        self.positive_ratio = positive_ratio
        self.sx_array   = None
        self.anomalies  = None

    def estimate_rg( self, interval, qrg_limits=None, Rg=None ):
        if Rg is None or qrg_limits is None:
            f_, t_ = interval
        else:
            f_, t_ = self.get_qrg_limits_indeces( interval, qrg_limits, Rg )

        # temp fix
        if t_ >= len( self.x ):
            t_ = len( self.x ) - 1

        if DEBUG: print( 'estimate_rg: (1) f_, t_=', ( f_, t_ ) )
        if f_ >= t_:
            raise Exception( 'f_(%d) >= t_(%d)' % ( f_, t_  ) )

        slice_ = slice( f_, t_+1 )
        x = self.x[ slice_ ]
        y = self.y[ slice_ ]
        e = self.e[ slice_ ]

        w = 1/e**2
        wls = WLS_Rg( x, y, w )

        self.wls        = wls
        self.interval   = [ f_, t_ ]
        if DEBUG: print( 'estimate_rg: (2) ', (f_, t_), self.wls.Rg )

    def get_qrg_limits_indeces( self, interval, qrg_limits, Rg  ):
        f_, t_ = interval

        limits = list( map( lambda q: bisect_right( self.x, q/Rg ), qrg_limits ) )
        if limits[0] > f_ and limits[0] < t_:
            f_ = limits[0]
        if limits[1] < t_ and limits[1] > f_:
            t_ = limits[1]

        if DEBUG: print( interval, '=>', (f_, t_) )

        return f_, t_

    def get_quality( self, fit ):
        # TODO: interval conversion
        if self.wls is None:
            quality_obj = Quality( fit, None, None, self.interval, self.positive_ratio, self.x )
        else:
            quality_obj = Quality( fit, self.wls.Rg, self.wls.sigmaRg, self.interval, self.positive_ratio, self.x )
        return quality_obj

    def get_basic_result( self, fit=None ):
        return (
                self.wls.B, self.wls.A,
                self.wls.Rg, self.wls.sigmaRg,
                self.wls.I0, self.wls.sigmaI0,
                )

    def get_result( self, fit=None, result_type=None, result=None ):

        self.fit = fit

        if DEBUG: print( 'get_result----------------------------------------------------------' )

        if result is None:
            if DEBUG: print( 'Guinier: get_result None' )

            if self.wls is None:
                Rg          = None
                Rg_stdev    = None
                I0          = None
                I0_stdev    = None
                a, b        = None, None
            else:
                wls         = self.wls
                Rg          = wls.Rg
                Rg_stdev    = wls.sigmaRg
                I0          = wls.I0
                I0_stdev    = wls.sigmaI0
                a, b        = wls.B, wls.A

            if self.interval is None:
                f_, t_      = None, None
                minQ        = None
                maxQ        = None
            else:
                f_, t_      = self.interval
                minQ        = np.sqrt( self.x[f_] )
                maxQ        = np.sqrt( self.x[t_] )

        else:
            if DEBUG: print( 'Guinier: get_result from Struct' )

            assert( hasattr( result, 'wls' ) )
            self.wls    = result.wls
            wls         = result.wls
            Rg          = wls.Rg
            Rg_stdev    = wls.sigmaRg
            I0          = wls.I0
            I0_stdev    = wls.sigmaI0
            a, b        = wls.B, wls.A

            f_, t_      = result.From, result.To
            self.interval = [ f_, t_ ]
            minQ        = np.sqrt( self.x[f_] )
            maxQ        = np.sqrt( self.x[t_] )

            if result_type is None:
                result_type = result.result_type

        curvatures_array = None
        IpI             = None
        head_cvmin      = None
        bicomponent  = None
        bico_mono_ratio    = None
        anomalies       = None

        if fit is None or self.wls is None:
            quality_obj     = None
            quality         = None
            quality_signal  = None
            head_trend      = None
            anomaly_index   = 0
            head_trend      = None
            head_proportion = None
            remarks         = None
        else:
            remarks         = fit.remarks

            quality_obj     = self.get_quality( fit )

            head_trend, head_proportion = self.compute_head_data_trend( fit )
            if DEBUG: print( 'head_trend, head_proportion=', head_trend, head_proportion )

            try:
                anomalies   = self.compute_anomaly_indeces( fit, quality_obj, minQ, I0, Rg, head_trend )
                if anomalies is not None:
                    IpI             = anomalies[2]
                    head_cvmin      = anomalies[3]
                    bicomponent  = anomalies[4]      # not yet set
                    bico_mono_ratio    = anomalies[5]      # not yet set
            except:
                etb = ExceptionTracebacker()
                print(etb)

            quality         = quality_obj.quality
            quality_signal  = quality_obj.quality_signal

        if fit is None:
            gpfit_RG    = None
            gpfit_I0    = None
            gpfit_d     = None
            gpfit_minQ  = None
            gpfit_maxQ  = None
        else:
            gpfit_RG    = fit.Rg
            gpfit_I0    = fit.I0
            gpfit_d     = fit.degree
            # gpfit_minQ  = fit.Q[fit.f0]
            # gpfit_maxQ  = fit.Q[fit.t0]
            gpfit_minQ  = fit.Q[0]
            gpfit_maxQ  = fit.Q[-1]

        if quality_obj is None:
            basic_quality = None
            positive_score  = None
            fit_cover_ratio = None
            fit_consistency = None
            stdev_score     = None
            q_rg_score      = None
        else:
            basic_quality = quality_obj.basic_quality
            positive_score  = quality_obj.positive_score
            fit_cover_ratio = quality_obj.fit_cover_ratio
            fit_consistency = quality_obj.fit_consistency
            stdev_score     = quality_obj.stdev_score
            q_rg_score      = quality_obj.q_rg_score

        if anomalies is None or anomalies[-1] is None:
            bico_G1, bico_G2, bico_Rg1, bico_Rg2, bico_d1, bico_d2 = [ None ] * 6
        else:
            bico_G1, bico_G2, bico_Rg1, bico_Rg2, bico_d1, bico_d2 = anomalies[-1] 

        if DEBUG: print( 'bicomponent=', bicomponent )

        result = Result(
                    type='E',
                    Rg=Rg, Rg_stdev=Rg_stdev,
                    I0=I0, I0_stdev=I0_stdev,
                    From=f_, To=t_,
                    Quality=quality,
                    Aggregated=head_trend,
                    IpI=IpI,
                    head_cvmin=head_cvmin,
                    bicomponent=bicomponent,
                    bico_mono_ratio=bico_mono_ratio,
                    min_qRg=None if Rg is None else minQ * Rg,
                    max_qRg=None if Rg is None else maxQ * Rg,
                    min_curvature=None,
                    max_curvature=None,
                    a=a,
                    b=b,
                    remarks=remarks,
                    fit=fit,

                    gpfit_Rg    = gpfit_RG,
                    gpfit_I0    = gpfit_I0,
                    gpfit_d     = gpfit_d,
                    gpfit_minQ  = gpfit_minQ,
                    gpfit_maxQ  = gpfit_maxQ,

                    basic_quality = basic_quality,
                    positive_score  = positive_score,
                    fit_cover_ratio = fit_cover_ratio,
                    fit_consistency = fit_consistency,
                    fit_consistency_pure = None if quality_obj is None else quality_obj.fit_consistency_pure,
                    stdev_score     = stdev_score,
                    q_rg_score      = q_rg_score,

                    bico_G1  = bico_G1,
                    bico_G2  = bico_G2,
                    bico_Rg1 = bico_Rg1,
                    bico_Rg2 = bico_Rg2,
                    bico_d1  = bico_d1,
                    bico_d2  = bico_d2,
                    bico_result = None,

                    signal=quality_signal,
                    min_q=minQ,
                    max_q=maxQ,

                    # quality details
                    quality_object=quality_obj,
                    head_trend=head_trend,
                    head_proportion=head_proportion,
                    curvatures_array = curvatures_array,
                    result_type=result_type,
                    anomalies=anomalies,
                    )

        if DEBUG:
            fit_consistency_pure = None if quality_obj is None else quality_obj.fit_consistency_pure
            print( [f_, t_], ' Rg=', Rg, ', Quality=', quality, ', fit_pure=', fit_consistency_pure )

        return result

    def update_result_with_bico_solver( self, result ):

        if DEBUG: print( 'update_result_with_bico_solver: result.anomalies=', result.anomalies )

        bicomponent, bico_mono_ratio, bico_result = self.compute_bicomponent_info( result )
        # print( 'bicomponent=', bicomponent, '; bico_mono_ratio=', bico_mono_ratio )

        result.bicomponent    = bicomponent
        result.bico_mono_ratio  = bico_mono_ratio
        result.bico_result      = bico_result

        if bico_result is not None:
            G1, G2, Rg1, Rg2, d1, d2 = bico_result[1]
            G_ = G1 + G2
            result.bico_G1  = int(G1/G_*100)
            result.bico_G2  = int(G2/G_*100)
            result.bico_Rg1 = Rg1
            result.bico_Rg2 = Rg2
            result.bico_d1  = d1
            result.bico_d2  = d2

        return result

    def compute_head_data_trend( self, fit ):
        if self.interval is None:
            return None, None

        # f0_ = fit.start_point
        f0_      = fit.f0
        f2_, t2_ = self.interval

        head_size = f2_ - f0_ + 1

        if DEBUG: print( 'head_size=', head_size, ', self.wls.sigmaB=', self.wls.sigmaB )

        if head_size >= 8:
            # avoid calculation for too short intervals
            # print( 'head trend: f0_, f2_=', (f0_, f2_) )
            slice_ = slice( f0_, f2_ )
            x_ = self.x[slice_]
            y_ = self.y[slice_]
            w_ = 1/self.e[slice_]**2

            ht = WLS_HeadTrend( x_, y_, w_ )
            guin_size = self.x[t2_] - self.x[f2_]
            trend, prop = ht.compute_trend( self.wls.B, guin_size )

        else:
            trend, prop = 0, 1

        return trend, prop

    def compute_anomaly_indeces( self, fit, quality_obj, minQ, I0, Rg, head_trend ):

        # f0_ = fit.start_point
        f0_ = fit.f0

        t_array = []

        """"
        if quality_obj.fit_cover_ratio < 0.5:
            d2_ex = 1
        else:
            d2_ex = 1.5 if fit.degree >= 3 else 2
            # d2_ex = 1.5
        d_params = [ [ 1, 1 ], [ fit.degree, d2_ex ] ]
        if quality_obj.fit_consistency_pure < 0.5 or quality_obj.fit_cover_ratio < 0.5:
            # if the consistency is low, check also in a wider interval
            d4_ex = 1.5 if fit.degree >= 3 else 2       # if fit.degree is small, you must make the interval wider.
            d_params.append( [ 4, d4_ex ] )
        """

        for d, ex in [ [ 1, 0.5 ] ]:
            q_ = 1/fit.Rg * np.sqrt( 3*d/2 )
            qq_ = q_**2 * ex
            """
                ex=1.5 for d=2 is too narrow    for subtract_SSC
                ex=2   for d=4 is too wide      for OA
            """
            t_ = bisect_right( self.x, qq_ )
            t_array.append( [ d, t_ ] )

        self.sx_array = []
        self.spline_array = []
        self._y_array = []
        self.sy_array = []
        self.cy_array = []
        self.curvatures_array = []

        if DEBUG: print( 'compute_anomaly_indeces: t_array=', t_array, ', y.shape[0]=', self.y.shape[0] )

        for i, param in enumerate( t_array ):

            d, t = param
            t_ = max( f0_ + 10, t )

            if t_ >= self.y.shape[0]:
                t_ = self.y.shape[0] - 1

            if DEBUG: print( '[%d] d=%g, f0_=%d, t_=%d' % ( i, d, f0_, t_ ) )

            slice_ = slice( f0_, t_+1 )

            sx = self.x[slice_]
            _y = self.y[slice_]
            w_ = 1/self.e[slice_]

            try:
                # with_smoother = i == 0
                with_smoother = False       # smoother not required by IPI_MAX_STDEV_RATIO check
                smooth_spline = SmoothSpline( sx, _y, w=w_, curvature=True, with_smoother=with_smoother )
            except Exception as e:
                if DEBUG:
                    etb = ExceptionTracebacker()
                    print(etb)

                self.sx_array = None
                raise e

            spline  = smooth_spline.spline
            self.spline_array.append( spline )
            sy = spline( sx )
            cy = smooth_spline.cy
            curvatures = smooth_spline.cv_array

            self.sx_array.append( sx )
            self._y_array.append( _y )
            self.sy_array.append( sy )
            self.cy_array.append( cy )
            self.curvatures_array.append( curvatures )

        # self.sx = self.sx_array[0]
        # self._y = self._y_array[0]

        cvmin   = self.curvatures_array[0][0]

        if DEBUG:
            print(  'basic_quality=', quality_obj.basic_quality,
                    ', stdev_ratio=', quality_obj.stdev_ratio,
                    ', quality=', quality_obj.quality )

        if ( quality_obj.aic_score >= AIC_SCORE_BOUNDARY    # quality_obj.aic_score < AIC_SCORE_BOUNDARY for agg1 or IpI
            and (   quality_obj.basic_quality   < REQUIRED_BASIC_QUALITY
                or  quality_obj.quality         < REQUIRED_QUALITY
                )
            ):
            if DEBUG: print( "Can't analize anomalies." )
            return None

        if DEBUG: print( 'cvmin=', cvmin, '; curvatures_array=', self.curvatures_array )

        # ipi_value   = 0 if Rg is None or quality_obj.fit_cover_ratio > 0.5 else cvmin
        ipi_value   = cvmin
        if (  ( quality_obj.stdev_ratio <= IPI_MAX_STDEV_RATIO
                # and quality_obj.basic_quality >= IPI_REQ_BASIC_QUALITY
                # cvmin == -18 for AIMdim01_00199_sub.dat
                # or fit.intensity.ha_end >= 10
              )
                and ipi_value < IPI_CURVATURE_BOUNDARY ):
            ipi_index   = 1
        else:
            ipi_index   = 0

        if self.wls.Rg is not None:
            stdev_ratio = self.wls.sigmaRg / self.wls.Rg
        else:
            stdev_ratio = 1

        bico_mono_ratio   = None
        bico_params  = None

        if DEBUG: print(    'head_trend=', head_trend,
                            ', fit_consistency_pure=', quality_obj.fit_consistency_pure,
                            ', q_rg_score=', quality_obj.q_rg_score )

        if ( ( quality_obj.basic_condition >= POLY_REQ_BASIC_CONDITION
                or quality_obj.basic_condition >= POLY_REQ_BASIC_CONDITION - 0.1
                    and head_trend < HEAD_TREND_POLY_BOUNDARY    # relaxed for agg2/TYINAs105_ave_sub.dat
              )
            and ( quality_obj.fit_consistency_pure < FIT_CONSISTENCY_PURE_BOUNDARY
                or quality_obj.aic_score < AIC_SCORE_BOUNDARY
                or head_trend < HEAD_TREND_POLY_BOUNDARY
                or quality_obj.q_rg_score < POLY_QRG_SCORE_BOUNDARY  # for TY03s304_ave0-4_sub.dat, TY30s305_ave0-4_sub.dat
                or False     # True for bico_mono_ratio's for all
                )
            ):

            aggregation = None
            quality_obj.needs_bico_solver = True
            if head_trend < HEAD_TREND_POLY_BOUNDARY:
                aggregation = -1
        else:
            aggregation = 0

        # anomalies = [ aggregation, ipi_index, bico_mono_ratio, bico_params ]
        num_anomalous_points = fit.intensity.ha_end
        anomalies = [ head_trend, num_anomalous_points, ipi_index, cvmin, aggregation, bico_mono_ratio, bico_params ]

        if DEBUG: print( 'anomalies=', anomalies )

        self.anomalies = anomalies

        return anomalies

    def comupute_bico_mono_ratio( self, fit ):

        bico_result = None

        solver = BiComponentSolver( fit.intensity, no_assert=True )
        try:
            bico_result = solver.solve( fit=fit )
        except:
            etb = ExceptionTracebacker()
            print(etb)
            return None, bico_result

        bico_ee, bico_params, I0_est = bico_result

        G1, G2, Rg1, Rg2, d1, d2 = bico_params

        G_  = G1 + G2
        Rg_ = ( G1*Rg1 + G2*Rg2 ) / G_
        d_  = ( G1*d1  + G2*d2 )  / G_
        gp_ee_a = solver.evaluate_gp( G_, Rg_, d_ )
        gp_ee_f = solver.evaluate_gp( fit.I0, fit.Rg, fit.degree )
        gp_ee = min( gp_ee_a, gp_ee_f )

        bico_mono_ratio = bico_ee / gp_ee
        if DEBUG:
            print( 'Guinier: gp_ee_a=', gp_ee_a, ', gp_ee_f', gp_ee_f )
            print( 'Guinier: bico_ee=', bico_ee )
            print( 'Guinier: gp_ee=',  gp_ee )
            print( 'Guinier: bico_mono_ratio=',  bico_mono_ratio )

        return bico_mono_ratio, bico_result

    def compute_bicomponent_info( self, result ):

        bico_result = None

        # print( 'compute_bicomponent_info: needs_bico_solver=', result.quality_object.needs_bico_solver  )
        if result.quality_object is not None and result.quality_object.needs_bico_solver:
            bico_mono_ratio, bico_result = self.comupute_bico_mono_ratio( result.fit )
        else:
            bico_mono_ratio = None

        head_trend      = result.head_trend
        head_proportion = result.head_proportion
        quality_obj     = result.quality_object
        anomalies       = result.anomalies

        if head_trend is None:
            bicomponent  = None
        else:
            if DEBUG:
                print( 'anomalies=', anomalies )
                print( 'fit_cover_ratio=', quality_obj.fit_cover_ratio, ', fit_score=', quality_obj.fit_score, ', head_trend=', head_trend )

            if quality_obj.basic_condition < REQUIRED_BASIC_QUALITY:

                bicomponent  = None

            else:
                # TODO: clarify the case where bico_mono_ratio is None

                polycomp_score = 0

                if quality_obj.stdev_ratio < ANALYZABLE_STDEV_RATIO:
                    if head_trend < HEAD_TREND_POLY_BOUNDARY and head_proportion > 0.5:
                        polycomp_score  += 1

                if bico_mono_ratio is not None:

                    if quality_obj.stdev_ratio < POLY_STRICT_STDEV_RATIO:
                        if bico_mono_ratio < POLY_WIDE_RATIO_BOUNDARY:
                            polycomp_score  += 1    # for ts008_ave_sub.dat
                    else:
                        if bico_mono_ratio < POLY_NARROW_RATIO_BOUNDARY:
                            polycomp_score  += 1    # except for AIMdim01_00226_sub.dat

                    G1, G2, Rg1, Rg2, d1, d2 = bico_result[1]
                    I0 = G1 + G2
                    pc1 = G1 / I0
                    pc2 = G2 / I0
                    Rg_diff = abs( Rg1 - Rg2 ) / ( Rg1 + Rg2 )
                    # print( 'Rg_diff=', Rg_diff )

                    if (    pc1 < POLY_ASSERTABLE_PROPORTION
                        and pc2 < POLY_ASSERTABLE_PROPORTION
                        and Rg_diff > POLY_ASSERTABLE_DIFF_RATIO
                       ):
                       polycomp_score  += 1

                # print( 'polycomp_score=', polycomp_score )
                if polycomp_score >= 2:
                    bicomponent   = 1
                else:
                    bicomponent   = 0

                if result.anomalies is not None:
                    result.bicomponent    = bicomponent
                    result.anomalies[5]     = bico_mono_ratio

        return bicomponent, bico_mono_ratio, bico_result

    def determine_ipi_interval( self ):
        # print( 'determine_ipi_interval' )
        gp_slope = - self.fit.Rg**2 / 3

        start_ipi_retry = None

        sx = self.sx_array[0]
        sy = self.sy_array[0]
        for i in range( len( sx ) - 10 ):
            x0 = sx[i]
            y0 = sy[i]
            x1 = sx[i+10]
            y1 = sy[i+10]
            slope = ( y1 - y0 ) / ( x1 - x0 )
            slope_ratio = slope/gp_slope
            # print( '[%d] slope ratio=%g' % ( i, slope_ratio ) )
            if slope_ratio >= 0.95:
                break

        # this includes the case when slope_ratio < 0.95
        start_ipi_retry = i

        return start_ipi_retry



"""
    Optimizer.ResultFolderSelector.py

    Copyright (c) 2021-2024, SAXS Team, KEK-PF
"""
import os
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry
from molass_legacy._MOLASS.SerialSettings import set_setting
from molass_legacy._MOLASS.Version import is_developing_version
from .FullOptResult import FILES
from .FullOptUtils import in_folder_consistency_ok

def job_folder_checker(self):
    optjob_folder = self.folder.get()
    files_ok = True
    for file in FILES[:-1]: # except "bounds.txt", which is optional
        path = os.path.join(optjob_folder, file)
        if not os.path.exists(path):
            files_ok = False
    entry = self.folder_entry
    if files_ok:
        path = os.path.join(optjob_folder, FILES[1])    # "in_data_info.txt"
        if in_folder_consistency_ok(self.in_folder, path, self):
            entry.config(fg="black")
            self.proceed_btn.config(state=Tk.NORMAL)
        else:
            entry.config(fg="red")
    else:
        entry.config(fg="red")
        self.update()
        import molass_legacy.KekLib.CustomMessageBox as MessageBox
        MessageBox.showerror("Invalid Folder",
            'This folder does not include required files.\n'
            'Select a folder named like "*/jobs/nnn".',
            parent=self)

def job_folder_applier(self):
    folder = self.folder.get()
    set_setting("optjob_folder", folder)
    optimizer_folder = os.path.dirname(os.path.dirname(folder))
    set_setting("optimizer_folder", optimizer_folder)

def joblist_folder_checker(self):
    # TODO: check if the folder is a job list folder
    folderpath = self.folder.get()
    print("joblist_folder_checker: folder=", folderpath)
    foldername = folderpath.split("/")[-1]
    if not foldername == "jobs":
        self.folder_entry.config(fg="red")
        self.update()
        import molass_legacy.KekLib.CustomMessageBox as MessageBox
        MessageBox.showerror("Invalid Folder",
            'This folder does not end with "jobs".\n'
            'Select a folder named like "*/jobs".',
            parent=self)
    else:
        self.folder_entry.config(fg="black")
        self.proceed_btn.config(state=Tk.NORMAL)

def joblist_folder_applier(self):
    folder = self.folder.get()
    optimizer_folder = os.path.dirname(folder)
    set_setting("optimizer_folder", optimizer_folder)

TARGET_FOLDER_DICT = {
    "Result Viewer" : ("Job Folder", job_folder_checker, job_folder_applier),
    "Result Animation" : ("Job List Folder", joblist_folder_checker, joblist_folder_applier),
}

if is_developing_version():
    TARGET_FOLDER_DICT["Parameter Transition"] = ("Job List Folder", joblist_folder_checker, joblist_folder_applier)

class ResultFolderSelector(Dialog):
    def __init__(self, parent, in_folder, proc_name):
        self.parent = parent
        self.in_folder = in_folder
        self.proc_name = proc_name
        self.applied = False
        Dialog.__init__(self, parent, proc_name, visible=False, location='lower center')

    def show(self):
        self._show()

    def body(self, body_frame):

        frame = Tk.Frame(body_frame)
        frame.pack(padx=20, pady=10)

        prompt_name, checker, applier = TARGET_FOLDER_DICT[self.proc_name]
        self.applier = applier

        label = Tk.Label(frame, text=prompt_name + ":")
        label.pack(side=Tk.LEFT)

        self.folder = Tk.StringVar()
        self.folder_entry = FolderEntry(frame, textvariable=self.folder, width=80,
            on_entry_cb=lambda: checker(self))
        self.folder_entry.pack(side=Tk.LEFT)

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=20, pady=5)
        w = Tk.Button(box, text="Proceed", width=10, command=self.ok, state=Tk.DISABLED)
        w.pack(side=Tk.LEFT, padx=20, pady=5)
        self.proceed_btn = w

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def apply(self):
        from molass_legacy._MOLASS.SerialSettings import set_setting
        self.applier(self)
        self.applied = True

    def get_logger(self):
        from molass_legacy.KekLib.ChangeableLogger import Logger
        logger = Logger(os.path.join(self.folder.get(), "opt_tracer.log"))
        return logger

    def get_result(self, folder=None, debug=False):
        if debug:
            import Optimizer.FullOptResult
            from importlib import reload
            reload(Optimizer.FullOptResult)
        from .FullOptResult import FullOptResult
        if folder is None:
            folder = self.folder.get()
        result = FullOptResult(self.parent.serial_data, self.parent.pre_recog, folder)
        return result

def result_viewer_impl(selector_dialog, parent_dialog, debug=False):
    if debug:
        from importlib import reload
        import Optimizer.FullOptDialog
        reload(Optimizer.FullOptDialog)
    from .FullOptDialog import FullOptDialog
    from molass_legacy.Trimming import get_trimming_info_list, set_trimming_info_list

    save_list = get_trimming_info_list()

    optjob_folder = selector_dialog.folder.get()
    set_setting('optworking_folder', optjob_folder)
    logger = selector_dialog.get_logger()
    logger.info("preparing FullOptDialog for view from %s.", optjob_folder)

    result = selector_dialog.get_result()
    optinit_info = result.get_init_info()   # this should restore required temporary settings by settings.load()

    demo_index = result.get_demo_index()
    job_name = result.get_job_name()
    dialog = FullOptDialog(parent_dialog.parent, parent_dialog, optinit_info, demo_index=demo_index, inspecting=True, job_name=job_name)
    parent_dialog.fullopt_dialog = dialog
    dialog.show()

    set_trimming_info_list(save_list, logger)

def result_animation_bridge(selector_dialog, debug=False):
    if debug:
        from importlib import reload
        import Optimizer.ResultAnimation
        reload(Optimizer.ResultAnimation)
    from molass_legacy.Optimizer.ResultAnimation import show_result_animation_impl
    show_result_animation_impl(selector_dialog)

def parameter_transition_brigde(selector_dialog, debug=False):
    if debug:
        from importlib import reload
        import Optimizer.ParameterTransition
        reload(Optimizer.ParameterTransition)
    from molass_legacy.Optimizer.ParameterTransition import show_parameter_transition_impl
    try:
        show_parameter_transition_impl(selector_dialog)
    except:
        import traceback
        traceback.print_exc()

def show_result_folder_selector_impl(parent_dialog, proc_name, ready_cb=None, debug=False):
    """
    task: consider unifying with Optimizer.OptimizerUtils.show_peak_editor_impl
    """

    in_folder = parent_dialog.in_folder.get()
    selector_dialog = ResultFolderSelector(parent_dialog, in_folder, proc_name)
    if ready_cb is not None:
        ready_cb(selector_dialog)

    parent_dialog.fullopt_dialog = None
    selector_dialog.show()

    IMPL_CLOSURE_DICT = {
        "Result Viewer": lambda: result_viewer_impl(selector_dialog, parent_dialog, debug=debug),
        "Result Animation" : lambda: result_animation_bridge(selector_dialog, debug=debug),
    }

    if is_developing_version():
        IMPL_CLOSURE_DICT["Parameter Transition"] = lambda: parameter_transition_brigde(selector_dialog, debug=True)

    if selector_dialog.applied:
        impl_closure = IMPL_CLOSURE_DICT[proc_name]
        impl_closure()


import pyPRISM
import matplotlib.pyplot as plt

# The system holds all information needed to set up a PRISM problem. We
# instantiate the system by specifying the site types and thermal energy
# level (kT, coarse-grained temperature) of the system.
sys = pyPRISM.System(['particle','polymer'],kT=1.0)

# We must discretize Real and Fourier space
sys.domain = pyPRISM.Domain(dr=0.01,length=4096)

# The composition of the system is desribed via number densities
sys.density['polymer']  = 0.75
sys.density['particle'] = 6e-6

# The diameter of each site is specified (in reduced units)
sys.diameter['polymer']  = 1.0
sys.diameter['particle'] = 5.0

# The molecular structure is described via intra-molecular correlation
# functions (i.e. omegas)
sys.omega['polymer','polymer']   = pyPRISM.omega.FreelyJointedChain(length=100,l=4.0/3.0)
sys.omega['polymer','particle']  = pyPRISM.omega.NoIntra()
sys.omega['particle','particle'] = pyPRISM.omega.SingleSite()

# The site-site interactions are specified via classes which are lazily
# evaluated during the PRISM-object creation
sys.potential['polymer','polymer']   = pyPRISM.potential.HardSphere(sigma=1.0)
sys.potential['polymer','particle']  = pyPRISM.potential.Exponential(sigma=3.0,alpha=0.5,epsilon=1.0)
sys.potential['particle','particle'] = pyPRISM.potential.HardSphere(sigma=5.0)

# Closure approximations are also specified via classes
sys.closure['polymer','polymer']   = pyPRISM.closure.PercusYevick()
sys.closure['polymer','particle']  = pyPRISM.closure.PercusYevick()
sys.closure['particle','particle'] = pyPRISM.closure.HyperNettedChain()

# Calling the .solve() method of the system object attempts to numerically
# solv the PRISM equation and, if successful, it returns a PRISM object
# containing all of the solved correlation functions.
PRISM = sys.solve()

# Calculate the pair-correlation functions.
rdf = pyPRISM.calculate.pair_correlation(PRISM)

# Plot the results using matplotlib
plt.plot(sys.domain.r,rdf['polymer','polymer'],color='gold',lw=1.25,ls='-')
plt.plot(sys.domain.r,rdf['polymer','particle'],color='red',lw=1.25,ls='--')
plt.plot(sys.domain.r,rdf['particle','particle'],color='blue',lw=1.25,ls=':')
plt.ylabel('pair correlation')
plt.xlabel('separation distance')
plt.show()



"""
https://arxiv.org/pdf/2103.16944.pdf

A CONCISE INTRODUCTION TO MOLECULAR DYNAMICS
    SIMULATION: THEORY AND PROGRAMMING

Ashkan Shekaari, Mahmoud Jafari
April 1, 2021

code1.py
"""
import numpy as np
from math import *
import random
__d__ = 3
NS = 1000
N = 100
dt = 1.e-4
T_0 = 300.
sig = 1.
eps = 1.
r_ctf = 2.5* sig
u_at_ctf = 4.* eps *(( sig / r_ctf ) **12 - ( sig/ r_ctf ) **6)
du_at_ctf = 24.* eps *( -2.* sig **12/ r_ctf **13 \
                    + sig **6/ r_ctf **7)
bs = 10.* sig
vol = bs **3
rho = N/ vol
ign = 20
#-------------------------------------------------
# Initiate positions, velocities, & accelerations
pos = np.zeros([__d__, N])
vel = np.zeros([__d__, N])
acc = np.zeros([__d__, N])
for j in range (N):
    for i in range (__d__):
        pos[i,j] = random.uniform(0, bs)
        vel[i,j] = random.uniform(0, 1)
#--------------------------------------
# Scale positions to the box size .
pos = pos/bs
#----------------------------
# Translate positions to the
# center -of - mass ( com) reference frame .
com = np . zeros ([ __d__ ])
for i in range ( __d__ ):
    for j in range (N):
        com [i] += pos[i , j]
        com [i] = com [i ]/ N
for i in range ( __d__ ):
    pos [i , :] -= com [i]
# - - - - - - - - -- ---- --- ----
P_S = 0.
k_S = 0.
p_S = 0.
T_S = 0.
f_tpz = open ('tpz. out', 'w')
f_kpe = open ('kpe. out', 'w')
f_xyz = open ('pos. xyz', 'w')
R = np . zeros ([ __d__ , N , N ])
r = np . zeros ([ __d__ , N , N ])
# - - - - - - - - -- ---- --- ---- --- --
def sign (a , b):
    """
    The sign function
    """
    if b > 0:
        return a
    elif b < 0:
        return -a
    elif b == 0:
        return a
# - - - - - - - - -- ---- --- ---- --- --- ---- --- --
# The time evolution loop ( main loop )
for step in range (1 , NS + 1) :
    # - - - - - - - - --- --- ---- --- --- ---- -
    # Periodic boundary conditions
    pos [ np . where ( pos > 0.5) ] -= 1
    pos [ np . where ( pos < -0.5) ] += 1
    # - - - - - - - - --- --- ---- --- --- ---- -
    # Loop for computing forces
    acc = np . zeros ([ __d__ , N ])
    pot = np . zeros ([ N ])
    vrl = 0.
    for i in range ( N):
        for j in range (i + 1 , N) :
            if j != i:
                r2 = np . zeros ([N , N ])
                for k in range ( __d__ ):
                    R[k , i , j] = pos [k , i] - pos[k , j]
                    if abs (R [k , i , j ]) > 0.5:
                        R[k , i , j ] -= sign (1. , R[k , i , j ])
                    r[k , i , j] = bs *R[k , i , j]
                    r2 [i , j] += r[k , i , j ]* r[k , i , j ]
            if r2 [i , j] < r_ctf * r_ctf :
                r1 = sqrt ( r2 [i , j ])
                ri2 = 1./ r2 [i , j ]
                ri6 = ri2 * ri2 * ri2
                ri12 = ri6 * ri6
                sig6 = sig **6
                sig12 = sig **12
                u = 4.* eps *( sig12 * ri12 - sig6 * ri6 ) \
                        - u_at_ctf - r1 * du_at_ctf
                du = 24.* eps * ri2 *(2.* sig12 * ri12 - sig6 * ri6) \
                        + du_at_ctf* sqrt ( ri2 )
                pot [j ] += u
                vrl -= du * r2 [i , j]
                for k in range ( __d__ ):
                    acc [k , i ] += du *R[k , i , j ]
                    acc [k , j ] -= du *R[k , i , j ]
    vrl = - vrl / __d__
    # - - - - - - - - --- --- ----
    # Update positions
    pos += dt * vel + 0.5* acc * dt * dt
    # - - - - - - - - --- --- ---- --- --- ----
    # Compute temperature
    kin = np . zeros ([ N ])
    v2 = np . zeros ([ N ])
    for j in range ( N):
        for i in range ( __d__ ) :
            v2[j ] += vel[i, j]*vel[i, j]* bs * bs
        kin[ j] = 0.5* v2 [j]
    k_AVG = sum( kin )/N
    T_i = 2.* k_AVG / __d__
    B = sqrt ( T_0 / T_i )
    # - - - - - - - - --- --- ---- --- --- ---- --- -
    # Rescale & update the velocities
    # according to velocity Verlet algorithm
    vel = B* vel + 0.5* dt * acc
    vel += 0.5* dt * acc
    # - - - - - - - - --- --- ---- ---
    # Compute temperature
    kin = np . zeros ([ N ])
    v2 = np . zeros ([ N ])
    for j in range ( N):
        for i in range ( __d__ ) :
            v2 [j ] += vel [i , j ]* vel [i , j ]* bs * bs
        kin[ j] = 0.5* v2 [j]
    k_AVG = sum( kin )/N
    T_i = 2.* k_AVG / __d__
    B = sqrt ( T_0 / T_i )
    # - - - - - - - - --- --- ----
    p_AVG = sum( pot )/N
    etot_AVG = k_AVG + p_AVG
    P = rho* T_i + vrl / vol
    Z = P * vol /( N * T_i )
    f_tpz . write ("%e %e % e %e\n " \
    %( step *dt , T_i , P , Z))
    f_kpe . write ("%e %e % e %e\n " \
    %( step *dt , k_AVG , p_AVG , etot_AVG))
    # - - - - - - - - --- --- ---- --- --- ---- --- ---
    # Write position components
    f_xyz . write ("%d \n\n " %( step ))
    for i in range ( N):
        f_xyz . write ("%e % e %e\n " %( pos [0 , i ]* bs , \
        pos [1 , i ]* bs , pos [2 , i ]* bs ))
    # - - - - - - - - --- --- ---- --- --- ---- --- -
    if step > ign :
        P_S += P
        k_S += k_AVG
        p_S += p_AVG
        T_S += T_i
    # - - - - - - - - -- ---- --- ---- ---
    if step %( NS *0.1) == 0: # To inform that the code is being run
        print (">>>>> Iteration # %d out of %d " %( step , NS ) )
print ('\n >>>>> Statistical Averages <<<<<\n')
print ('Temperature = %f K' %( T_S /( NS - ign )) )
print ('Kinetic energy = %f eps' %( k_S /( NS - ign )))
print ('Potential energy = %f eps' %( p_S /( NS - ign )))
print ('Total energy = %f eps' %(( k_S + p_S ) /( NS - ign )) )
print ('Pressure = %f eps /( sig ^3) \n' %( P_S /( NS - ign )))



"""
    Optimizer.FuncReloadUtils.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
from .FuncImporter import import_objective_function

def reload_optimizer(old_optimizer, prepare=True, init_params=None):
    func_code = old_optimizer.get_name()
    func_class = import_objective_function(func_code)

    optimizer = func_class(
                old_optimizer.dsets,
                old_optimizer.n_components,
                uv_base_curve=old_optimizer.uv_base_curve,
                xr_base_curve=old_optimizer.xr_base_curve,
                qvector=old_optimizer.qvector,    # trimmed sd
                wvector=old_optimizer.wvector,
                )

    if old_optimizer.is_stochastic():
        optimizer.params_type.set_estimator(old_optimizer.params_type.estimator)

    if prepare:
        if init_params is None:
            init_params = old_optimizer.init_params
        optimizer.prepare_for_optimization(init_params)

    return optimizer


"""
    ClipboardConstants.py

    from https://learn.microsoft.com/en-us/windows/win32/dataxchg/standard-clipboard-formats
"""

CF_BITMAP = 2
CF_DIB = 8
CF_DIBV5 = 17
CF_DIF = 5
CF_DSPBITMAP = 0x0082

CF_DSPENHMETAFILE = 0x008E
CF_DSPMETAFILEPICT = 0x0083
CF_DSPTEXT = 0x0081
CF_ENHMETAFILE = 14
CF_GDIOBJFIRST = 0x0300

CF_GDIOBJLAST = 0x03FF
CF_HDROP = 15
CF_LOCALE = 16
CF_METAFILEPICT = 3
CF_OEMTEXT = 7

CF_OWNERDISPLAY = 0x0080
CF_PALETTE = 9
CF_PENDATA = 10
CF_PRIVATEFIRST = 0x0200
CF_PRIVATELAST = 0x02FF

CF_RIFF = 11
CF_SYLK = 4
CF_TEXT = 1
CF_TIFF = 6
CF_UNICODETEXT = 13

CF_WAVE = 12

CF_ALL = [
    CF_BITMAP, CF_DIB, CF_DIBV5, CF_DIF, CF_DSPBITMAP,
    CF_DSPENHMETAFILE, CF_DSPMETAFILEPICT, CF_DSPTEXT, CF_ENHMETAFILE, CF_GDIOBJFIRST,
    CF_GDIOBJLAST, CF_HDROP, CF_LOCALE, CF_METAFILEPICT, CF_OEMTEXT,
    CF_OWNERDISPLAY, CF_PALETTE, CF_PENDATA, CF_PRIVATEFIRST, CF_PRIVATELAST,
    CF_RIFF, CF_SYLK, CF_TEXT, CF_TIFF, CF_UNICODETEXT,
    CF_WAVE,
    ]



"""
    Estimators.TestTools.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

def EGH_estimator_test_impl(editor):
    from importlib import reload
    import Estimators.EghEstimator
    reload(Estimators.EghEstimator)
    from Estimators.EghEstimator import EghEstimator
    print("estimator_test_impl")
    estimator = EghEstimator(editor)
    estimator.estimate_params(debug=True)

def SDM_estimator_test_impl(editor):
    from importlib import reload
    import Estimators.SdmEstimator
    reload(Estimators.SdmEstimator)
    from Estimators.SdmEstimator import SdmEstimator
    print("estimator_test_impl")
    estimator = SdmEstimator(editor)
    optimizer = editor.fullopt
    opt_name = optimizer.get_name()
    print("optimizer name=", opt_name)
    edm_available = opt_name[0:2] == "G2"   # 

    init_params = estimator.estimate_params(edm_available=edm_available, debug=True)
    if init_params is None:
        return

    class_code = 'G1100'
    if edm_available:
        # create SDM optimizer
        from molass_legacy.Optimizer.FuncImporter import import_objective_function
        optimizer_class = import_objective_function(class_code)
        optimizer = optimizer_class(
            editor.dsets,
            editor.get_n_components(),
            uv_base_curve=editor.base_curve_info[0],
            xr_base_curve=editor.baseline_objects[1],
            qvector=editor.sd.qvector,    # trimmed sd
            wvector=editor.sd.lvector,
            )

        optimizer.params_type.set_estimator(estimator)      # reconsider the neccesity of this line
        optimizer.prepare_for_optimization(init_params)

    lrf_info = optimizer.objective_func(init_params, return_lrf_info=True)

    def plot_components(ax, x, y, C, ty):
        ax.plot(x, y, label="data")
        for k, cy in enumerate(C):
            ax.plot(x, cy, ":", label="component-%d" % k)
        ax.plot(x, ty, ":", color="red", label="model total")

    with plt.Dp():
        fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(12,10))
        fig.suptitle("estimator_test_impl")
        ax1, ax2 = axes[0,0], axes[0,1]
        ax3, ax4 = axes[1,0], axes[1,1]
        plot_components(ax1, lrf_info.uv_x, lrf_info.uv_y, lrf_info.matrices[3], lrf_info.uv_ty)
        plot_components(ax2, lrf_info.x, lrf_info.y, lrf_info.matrices[1], lrf_info.xr_ty)
        plot_components(ax3, lrf_info.uv_x, lrf_info.uv_y, lrf_info.scaled_uv_cy_array, lrf_info.uv_ty)
        plot_components(ax4, lrf_info.x, lrf_info.y, lrf_info.scaled_xr_cy_array, lrf_info.xr_ty)
        fig.tight_layout()
        ret = plt.show()

    if ret and edm_available:
        # replace editor optimizer
        # task: move this code to PeakEditor
        from molass_legacy._MOLASS.SerialSettings import set_setting
        from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
        set_setting('elution_model', 1)
        editor.optimizer = optimizer
        editor.fullopt = optimizer
        text = "Elution Decomposition of %s with func=%s" % (get_in_folder(), class_code)
        editor.fig.suptitle(text, fontsize=20)
        editor.draw_scores(init_params=init_params, create_new_optimizer=False)
        editor.fullopt_class = optimizer_class
        editor.class_code = class_code

    return ret

def EDM_estimator_test_impl(editor):
    from importlib import reload
    import Estimators.EdmEstimator
    reload(Estimators.EdmEstimator)
    from Estimators.EdmEstimator import EdmEstimator
    n_components = editor.get_n_components()
    print("estimator_test_impl: n_components=", n_components)
    estimator = EdmEstimator(editor, n_components=n_components)
    estimator.estimate_params(debug=True)

def estimator_test_impl(editor):

    extra_button_specs = [
        ("EGH Edtimator", lambda: EGH_estimator_test_impl(editor)),
        ("SDM Edtimator", lambda: SDM_estimator_test_impl(editor)),
        ("EDM Edtimator", lambda: EDM_estimator_test_impl(editor)),
    ]

    with plt.Dp(extra_button_specs=extra_button_specs):
        fig, ax = plt.subplots()
        plt.show()


# coding: utf-8
"""
    GuPy.py

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF
"""
import numpy as np
import pyopencl as cl
from pyopencl.array import Array
import pyclblast

default_dtype = np.float32

ctx = None
queue = None

def get_default_platform():
    platforms = cl.get_platforms()
    return platforms[0]

def initialize():
    global queue, ctx
    # print( 'np.float32=', np.float32 )

    platform = get_default_platform()
    print('platform name:', platform.name)
    devs = platform.get_devices()
    for d in devs:
        print('device name:', d.name)
        device_type = d.type
        device_type_str = cl.device_type.to_string(device_type)
        print('device type:', device_type_str)
        print('global memory size: ', d.global_mem_size / (1024**3), 'GB')
        if device_type_str == 'GPU':
            break

    if False:
        ctx = cl.create_some_context()
    else:
        ctx = cl.Context(
            dev_type=device_type,
            properties=[(cl.context_properties.PLATFORM, platform)])
    queue = cl.CommandQueue(ctx)

def terminate():
    global queue
    queue.finish()

def dot(a, b):
    global queue
    m, j = a.shape
    k, n = b.shape
    assert j == k
    c = np.zeros((m, n), dtype=default_dtype)
    cla = Array(queue, a.shape, a.dtype)
    clb = Array(queue, b.shape, b.dtype)
    clc = Array(queue, c.shape, c.dtype)
    cla.set(a)
    clb.set(b)
    clc.set(c)
    pyclblast.gemm(queue, m, n, k, cla, clb, clc, a_ld=k, b_ld=n, c_ld=n)
    return clc.get()



# coding: utf-8
"""
    MmAnimator.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import time
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from .GammaVisualizer import GammaVisualizer

class MmAnimator:
    def __init__(self, fig, axes, ylims, xy_list, mm_list, datasets,
                    step_mode=False, show_extrapolated=False):
        self.fig = fig
        self.axes = axes[1,:]
        self.axes_anim = axes[2,:]
        self.ylims = ylims
        self.xy_list = xy_list
        self.mm_list = mm_list
        self.datasets = datasets
        self.step_mode = step_mode
        self.show_extrapolated = show_extrapolated
        self.prepare()

        self.num_frames = mm_list[0].max_iter + 1
        if step_mode:
            self.current_step = 0
        else:
            self.anim = animation.FuncAnimation(fig, self.animate, frames=self.num_frames, blit=True, init_func=self.reset, interval=100)

    def prepare(self):
        def add_number_text(ax):
            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
            tx = (xmin + xmax)/2
            ty = (ymin + ymax)/2
            text = ax.text(tx, ty, "_", ha='center', va='center', alpha=0.05, fontsize=180)
            return text

        artist_list = []
        for ax, ylim, xy, mm in zip(self.axes, self.ylims, self.xy_list, self.mm_list):
            ax.set_ylim(ylim)
            x, y = xy
            ax.plot(x, y, label='input data')
            text = add_number_text(ax)
            ys = mm.get_anim_components(x, y, 0)
            line, = ax.plot(x, ys[0], label='eghmm-fit')
            artist_rec = [text, line] + [ax.plot(x, y_, ':', label='component-%d' % k)[0] for k, y_ in enumerate(ys[1:])]
            artist_list.append( artist_rec )
            ax.legend()

        if self.show_extrapolated:
            curves_list = []
            for ano, (mm, xy, ax) in enumerate(zip(self.mm_list, self.xy_list, self.axes_anim)):
                C = mm.get_anim_C(*xy, -1)
                lines = self.datasets.draw_exprapolated(ano, ax, C)[0]
                curves_list.append(lines)
            self.curves_matrix = np.array(curves_list)
        else:
            self.gv_list = []
            for ax, mm in zip(self.axes_anim, self.mm_list):
                gv = GammaVisualizer()
                poly, verts = gv.draw_gamma_3d(ax, mm.gamma_array[1,:,:])
                self.gv_list.append(gv)

        self.artist_matrix = np.array(artist_list)

        # self.artists = np.hstack([self.artist_matrix.flatten(), np.array(self.poly_list)])
        self.artists = self.artist_matrix.flatten()
        if self.show_extrapolated:
            self.artists_ex = np.hstack([self.artists, self.curves_matrix.flatten()])

        if self.step_mode:
            self.visual = True
            self.current_step = 0
            self.animate(0)
        else:
            artists_ = self.artists_ex if self.show_extrapolated else self.artists
            for art in artists_:
                art.set_visible(False)
            self.visual = False

    def reset(self):
        if not self.visual:
            artists_ = self.artists_ex if self.show_extrapolated else self.artists
            for art in artists_:
                art.set_visible(True)
            self.visual = True
        return self.animate(0)

    def animate(self, n):
        for i, (ax, xy, mm) in enumerate(zip(self.axes, self.xy_list, self.mm_list)):
            x, y = xy
            ys = mm.get_anim_components(x, y, n)
            artists = self.artist_matrix[i,:]
            artists[0].set_text(str(n))
            for art, y_ in zip(artists[1:], ys):
                art.set_data(x, y_)
            if self.show_extrapolated:
                ax2d = self.axes_anim[i]
                lines = self.curves_matrix[i,:]
                C = mm.get_anim_C(*xy, n)
                self.datasets.update_exprapolated(i, lines, C)
            else:
                ax3d = self.axes_anim[i]
                ax3d.cla()
                gv = self.gv_list[i]
                gamma = mm.gamma_array[n,:,:]
                gv.draw_gamma_3d(ax3d, gamma)

        if self.show_extrapolated:
            return self.artists_ex
        else:
            return self.artists


    """
    for the following cf.
    stop / start / pause in python matplotlib animation
    https://stackoverflow.com/questions/16732379/stop-start-pause-in-python-matplotlib-animation
    """
    def stop(self):
        self.anim.event_source.stop()

    def pause(self):
        self.anim.event_source.stop()

    def start(self):
        self.anim.event_source.start()

    def step(self, steps):
        self.current_step = i = (self.current_step + steps) % self.num_frames
        self.animate(i)

    def save(self, file, **kwargs):
        self.anim.save(file, **kwargs)



"""
    DebugCushion.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import molass_legacy.KekLib.DebugPlot as plt

def debug_impl(debug_info):
    print("debug_impl2")
    smoother = debug_info[0]
    x = smoother.x
    y = smoother.y
    # print(x, y)

    with plt.Dp():
        fig, ax = plt.subplots()
        ax.plot(x, y)
        plt.show()



# coding: utf-8
"""

    OurTk.py

    python-tkdnd
    https://libraries.io/pypi/python-tkdnd

    pip install python-tkdnd

    Copyright (c) 2021-2022, Masatsuyo Takahashi, KEK-PF

"""
from tkinter import *
import tkinter.font as font
from tkinterDnD import Tk



"""
    Optimizer.Devel.DevMeasureInspect.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

def dev_measure_inspect(optimizer, work_params):
    from importlib import reload
    import Distance.DeviationMeasure
    reload(Distance.DeviationMeasure)
    from Distance.DeviationMeasure import feature_deviation
    import Distance.NormalizedRmsd
    reload(Distance.NormalizedRmsd)
    from Distance.NormalizedRmsd import normalized_rmsd

    print("dev_measure_inspect")
    lrf_info1 = optimizer.objective_func(work_params, return_lrf_info=True)
    temp_params = work_params.copy()
    temp_params[2] = 0.3
    lrf_info2 = optimizer.objective_func(temp_params, return_lrf_info=True)
    
    adjust_target = 1
    uv_adjust = adjust_target - np.log10(optimizer.uv_norm1)
    xr_adjust = adjust_target - np.log10(optimizer.xr_norm1)
    print("uv_norm1=", optimizer.uv_norm1)
    print("xr_norm1=", optimizer.xr_norm1)
    print("uv_adjust=", uv_adjust)
    print("xr_adjust=", xr_adjust)

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12, 5))
        dev1 = feature_deviation(lrf_info1.xr_ty, lrf_info1.y, optimizer.xr_curve.max_y, optimizer.xr_norm1, debug=True, fig_info=(fig,ax1))
        dev2 = feature_deviation(lrf_info2.xr_ty, lrf_info2.y, optimizer.xr_curve.max_y, optimizer.xr_norm1, debug=True, fig_info=(fig,ax2))
        print("dev1=", dev1)
        print("dev2=", dev2)
        dev3 = normalized_rmsd(lrf_info1.uv_ty, lrf_info1.y, adjust=uv_adjust)
        dev4 = normalized_rmsd(lrf_info1.xr_ty, lrf_info1.y, adjust=xr_adjust)
        print("dev3=", dev3)
        print("dev4=", dev4)
        fig.tight_layout()
        plt.show()


"""
    SecTheory.SecCF.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import numpy as np

def simple_phi(w, npi, tpi):
    return np.exp(npi*(1/(1 - 1j*w*tpi) - 1))

def shifted_phi(w, npi, tpi, x0):
    return np.exp(npi*(1/(1 - 1j*w*tpi) - 1) + 1j*w*x0)

def moving_zone_phi(w, npi, tpi, N0, x0):
    Z = npi*(1/(1 - 1j*w*tpi) - 1) + 1j*w*x0
    return np.exp(Z + 1/(2*N0) * Z**2)

def two_site_phi(w, n, p1, p2, t1, t2):
    # see 1999, Alberto Cavazzini
    # return np.exp(n*p1/(1 - 1j*w*t1) - n*p1) * np.exp(n*p2/(1 - 1j*w*t2) - n*p2)
    return np.exp(n*(p1/(1 - 1j*w*t1) + p2/(1 - 1j*w*t2) -1))

"""
    GEC (Giddings-Eyring-Carmichael) models of SEC
"""
def gec_monopore_phi(w, npi, tpi, x0):
    # same as shifted_phi above
    return np.exp(npi*(1/(1 - 1j*w*tpi) - 1) + 1j*w*x0)

def gec_dipore_phi(w, n1, t1, n2, t2, x0):
    return np.exp(n1*(1/(1 - 1j*w*t1) - 1) + n2*(1/(1 - 1j*w*t2) - 1) + 1j*w*x0)

def gec_tripore_phi(w, n1, t1, n2, t2, n3, t3, x0):
    return np.exp(n1*(1/(1 - 1j*w*t1) - 1) + n2*(1/(1 - 1j*w*t2) - 1)+ n3*(1/(1 - 1j*w*t3) - 1) + 1j*w*x0)

def gec_polypore_phi(w, nt_pairs, x0):
    """
        this does not work properly. why?
        z = np.sum([n_*(1/(1 - 1j*w*t_) - 1) for n_, t_ in nt_pairs])
    """
    z = 0
    for n_, t_ in nt_pairs:
        z += n_*(1/(1 - 1j*w*t_) - 1)
    return np.exp(z + 1j*w*x0)

def gec_oligopore_phi_impl(w, nt_pairs, props, t0):
    z = 0
    for (n_, t_), p in zip(nt_pairs, props):
        z += p*n_*(1/(1 - 1j*w*t_) - 1)
    return np.exp(z + 1j*w*t0)

def sec_oligopore_phi_impl(w, nt_pairs, props, t0, N0):
    # see 2002, F. Dondi, formula (75)
    z = 0
    for (n_, t_), p in zip(nt_pairs, props):
        u = n_*(1/(1 - 1j*w*t_) - 1) + 1j*w*t0 
        z += p * (u + 1/(2*N0) * u**2)          # be careful in proportional scaling. is this correct?
    return np.exp(z )

"""
    Stochastic-dispersive models of SEC
"""
def sdm_monopore(w, npi, tpi, N0, t0, x0):
    # same as moving_zone_phi above
    Z = npi*(1/(1 - 1j*w*tpi) - 1) + 1j*w*t0
    return np.exp(Z + 1/(2*N0) * Z**2 + 1j*w*x0)

def sdm_dipore(w, np1, tp1, np2, tp2, N0, t0, x0):
    Z1 = np1*(1/(1 - 1j*w*tp1) - 1) + 1j*w*t0
    Z2 = np2*(1/(1 - 1j*w*tp2) - 1) + 1j*w*t0
    return np.exp((Z1 + 1/(2*N0) * Z1**2) + (Z2 + 1/(2*N0) * Z2**2) + 1j*w*x0)

def sdm_tripore(w, np1, tp1, np2, tp2, np3, tp3, N0, t0, x0):
    Z1 = np1*(1/(1 - 1j*w*tp1) - 1) + 1j*w*t0
    Z2 = np2*(1/(1 - 1j*w*tp2) - 1) + 1j*w*t0
    Z3 = np3*(1/(1 - 1j*w*tp3) - 1) + 1j*w*t0
    return np.exp((Z1 + 1/(2*N0) * Z1**2) + (Z2 + 1/(2*N0) * Z2**2) + (Z3 + 1/(2*N0) * Z3**2) + 1j*w*x0)

"""
    Stochastic-dispersive models of SEC
"""
def dispersive_monopore(w, npi, tpi, N0, t0):
    Z = npi*(1/(1 - 1j*w*tpi) - 1) + 1j*w*t0
    return np.exp(Z + Z**2/(2*N0))



"""
    Optimizer.OptLrfInfo.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

VARY_SMALL_VALUE = 1e-8

def get_ratio_cy_list(y, cy_list):
    basline = cy_list[-1]
    y_ = y - basline
    safe_ty = np.sum(cy_list[:-1], axis=0)   # excluding baseline
    where_small = np.abs(safe_ty) < VARY_SMALL_VALUE
    safe_ty[where_small] = VARY_SMALL_VALUE
    ret_list = []
    for cy in cy_list[:-1]:
        ret_list.append(y_*cy/safe_ty)
    ret_list.append(basline)    # basseline is not affected by ratio
    return ret_list

class OptLrfInfo:
    def __init__(self, Pxr, Cxr, Puv, Cuv, mapped_UvD,
                    qv, xrD, xrE,
                    x, y, xr_ty, scaled_xr_cy_array,
                    uv_x, uv_y, uv_ty, scaled_uv_cy_array,
                    composite):
        self.matrices = (Pxr, Cxr, Puv, Cuv, mapped_UvD)
        self.qv = qv
        self.xrD = xrD
        self.xrE = xrE
        self.x = x
        self.y = y
        self.xr_ty = xr_ty
        self.scaled_xr_cy_array = scaled_xr_cy_array
        self.uv_x = uv_x
        self.uv_y = uv_y
        self.uv_ty = uv_ty
        self.scaled_uv_cy_array = scaled_uv_cy_array
        self.composite = composite

    def get_num_substantial_components(self):
        return self.composite.get_num_substantial_components()

    def get_xr_cy_list(self):
        return list(self.scaled_xr_cy_array)

    def get_xr_proportions(self, exclude_baseline=True, debug=False):
        stop = -1 if exclude_baseline else None
        areas = []
        for cy in self.scaled_xr_cy_array[:stop]:
            areas.append(np.sum(cy))

        if debug:
            x = self.x
            y = self.y
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("get_xr_proportions")
                ax.plot(x, y)
                areas = []
                for cy in self.scaled_xr_cy_array:
                    ax.plot(x, cy, ":")
                    areas.append(np.sum(cy))
                areas = np.array(areas)
                props1 = areas/np.sum(areas)
                props2 = areas[:-1]/np.sum(areas[:-1])
                print(props1, props2)
                # with ALD002
                # [0.70246131 0.1551997  0.14233899] [0.81904307 0.18095693]
                xmin, xmax = ax.get_xlim()
                ymin, ymax = ax.get_ylim()
                tx = xmin*0.9 + xmax*0.1
                ty = ymin*0.3 + ymax*0.7
                dy = (ymax - ymin)*0.1
                ax.text(tx, ty, "props1=%s" % str(props1))
                ax.text(tx, ty-dy, "props2=%s" % str(props2))
                fig.tight_layout()
                plt.show()

        return np.array(areas)/np.sum(areas)

    def get_uv_cy_list(self):
        return list(self.scaled_uv_cy_array)

    def get_uv_proportions(self, exclude_baseline=True):
        stop = -1 if exclude_baseline else None
        areas = []
        for cy in self.scaled_uv_cy_array[:stop]:
            areas.append(np.sum(cy))
        return np.array(areas)/np.sum(areas)

    def estimate_xr_peak_points(self, elutions=None):
        if elutions is None:
            elutions = self.scaled_xr_cy_array
        return self.composite.estimate_peak_points(self.x, elutions)

    def estimate_uv_peak_points(self, elutions=None):
        if elutions is None:
            elutions = self.scaled_uv_cy_array
        return self.composite.estimate_peak_points(self.uv_x, elutions)

    def get_valid_rgs(self, rg_params):
        return self.composite.get_valid_rgs(rg_params)

    def get_scaled_xr_params(self, xr_params, debug=True):
        x = self.x
        ret_params = xr_params.copy()
        for k, m in enumerate(xr_params[:,1]):
            cy = self.scaled_xr_cy_array[k]
            ret_params[k,0] = cy[int(m - x[0])]

        if debug:
            from molass_legacy.Models.ElutionCurveModels import egh
            y = self.y
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("get_scaled_xr_params")
                ax.plot(x, y)
                cy_list = []
                for h, m, s, t in ret_params:
                    cy = egh(x, h, m, s, t)
                    cy_list.append(cy)
                    ax.plot(x, cy, ":")
                ty = np.sum(cy_list, axis=0)
                ax.plot(x, ty, ":", color="red")
                fig.tight_layout()
                plt.show()
        return ret_params

    def compute_rg_params(self, devel=True):
        if devel:
            from importlib import reload
            import Optimizer.GuinierRg
            reload(Optimizer.GuinierRg)
        from molass_legacy.Optimizer.GuinierRg import compute_rg_params_impl

        return compute_rg_params_impl(self)

    def update_optimizer(self, optimizer, params, devel=True):
        if devel:
            from importlib import reload
            import Optimizer.RgSecUpdater
            reload(Optimizer.RgSecUpdater)
        from molass_legacy.Optimizer.RgSecUpdater import update_optimizer_impl

        rg_params, rg_qualities = self.compute_rg_params()
        self.rg_params = rg_params
        self.rg_qualities = rg_qualities

        return update_optimizer_impl(optimizer, params, rg_params, rg_qualities)



"""

    StageSummary.py

    Copyright (c) 2016-2025, SAXS Team, KEK-PF

"""
import os
from openpyxl import Workbook
from molass_legacy.Reports.ReportUtils import make_summary_book
from molass_legacy.KekLib.ExcelCOM import merge_into_a_book
from molass_legacy.KekLib.ProgressInfo import put_info
from .ProgressInfoUtil import NUM_SHEET_TYPES, STREAM_ZERO_EX
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Reports.DefaultFont import set_default_font
set_default_font()

def do_summary_stage(self):
    do_summary(self)
    merge_books(self)

def do_summary( self ):
    # Summary Sheet
    summary_file = self.temp_folder + '/--summary.xlsx'

    if self.excel_is_available:
        wb = Workbook()
    else:
        wb = self.result_wb

    make_summary_book( self.excel_is_available, wb, summary_file, self )
    # self.temp_books.insert( 0, summary_file )
    self.temp_books.append( summary_file )
    self.logger.info( 'summary book done.' )

class MergeBooksArgs:
    def __init__(self, controller):
        self.book_path = os.path.abspath( controller.book_file )
        self.temp_books = controller.temp_books
        if controller.zx and controller.atsas_is_available:
            self.book_path_atsas = controller.atsas_folder + '/datgnom-result.xlsx'
            self.temp_books_atsas = controller.temp_books_atsas
        else:
            self.book_path_atsas = None
            self.temp_books_atsas = []

def merge_books(self, debug=False):
    # this is in order to avoid inconsistency in progressbar control
    # when then the notified number of steps becomes different
    # from the actual number of steps.
    max_num_progress = self.num_peaks_to_exec * NUM_SHEET_TYPES

    if self.excel_is_available:
        args = MergeBooksArgs(self)

        if self.more_multicore:
            self.teller.tell('merge_excel_books', args=args)
        else:
            def progress_callback(i):
                if i < max_num_progress:
                    if debug:
                        self.logger.info("put_info( (%d,1000), %d )", STREAM_ZERO_EX, i+1)
                    put_info( (STREAM_ZERO_EX,1000), i+1 )

            merge_books_impl(self.excel_client, args, self.logger, progress_callback)

        self.logger.info( 'books merge done.' )

    else:
        from molass_legacy.KekLib.OpenPyXlUtil import save_allowing_user_reply
        if len(self.temp_books) > 0:
            book_path = os.path.abspath( self.book_file )
            save_allowing_user_reply(self.result_wb, book_path)
        if len(self.temp_books_atsas) > 0:
            book_path_atsas = self.atsas_folder + '/datgnom-result.xlsx'
            save_allowing_user_reply(self.result_wb_datgnom, book_path_atsas)
        self.logger.info( 'books output complete.' )

    if debug:
        self.logger.info("put_info( (%d,1000), %d )", STREAM_ZERO_EX, max_num_progress)
    put_info( (STREAM_ZERO_EX,1000), max_num_progress )

def merge_books_impl(excel_client, args, logger, progress_cb=None):
    # logger is not yet used
    fontname = get_setting("report_default_font")

    if len(args.temp_books) > 0:
        merge_into_a_book( args.temp_books, args.book_path, excel_client=excel_client, progress_cb=progress_cb, delete_target="Sheet", default_font=fontname )

    if len(args.temp_books_atsas) > 0:
        merge_into_a_book( args.temp_books_atsas, args.book_path_atsas, excel_client=excel_client, progress_cb=progress_cb, default_font=fontname )



"""
    Models.Stochastic.Monopore.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.Models.Tentative import Model
from molass_legacy.SecTheory.BasicModels import robust_single_pore_pdf
from molass_legacy.Models.ElutionModelUtils import x_from_height_ratio_impl

def monopore_func(x, scale, N, T, t0, me, mp, poresize, rg):
    rho = min(1, rg/poresize)
    ni_ = N * (1 - rho)**me
    ti_ = T * (1 - rho)**mp
    return scale * robust_single_pore_pdf(x - t0, ni_, ti_)

class Monopore(Model):
    def __init__(self, **kwargs):
        super(Monopore, self).__init__(monopore_func, **kwargs)
        self.fx = None

    def get_name(self):
        return "STC"

    def is_traditional(self):
        return False

    def eval(self, params=None, x=None):
        return self.func(x, *params)
    
    def x_from_height_ratio(self, ecurve, ratio, params):
        assert self.fx is not None
        return x_from_height_ratio_impl(monopore_func, ecurve, ratio, *params, needs_ymax=True, full_params=True, fx=self.fx)
    
    def set_fx_for_height_ratio(self, fx):
        self.fx = fx

    def get_params_string(self, params):
        return 'scale=%g, N=%g, T=%g, t0=%g, me=%s, mp=%g, rp=%g, rg=%g' % tuple(params)


"""
    Optimizer.AllParameters.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
import re
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from ScrolledFrame import ScrolledFrame

class AllParameters(Dialog):
    def __init__(self, parent, optimizer, x_array, curr_index, best_index, x_, xmin, xmax):
        self.optimizer = optimizer
        self.x_array = x_array
        self.curr_index = curr_index
        self.best_index = best_index
        self.x_ = x_
        self.xlim = (xmin, xmax)
        Dialog.__init__(self, parent, "Variation of All Parameters", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = ScrolledFrame(body_frame)
        cframe.pack(side=Tk.LEFT)
        cframe.add_bind_mousewheel()

        cframe_ = Tk.Frame(cframe.interior)
        cframe_.pack()
        tframe_ = Tk.Frame(cframe.interior)
        tframe_.pack(side=Tk.LEFT)

        x_array = self.x_array
        nrows = x_array.shape[1]
        fig = plt.figure(figsize=(16, nrows*1))
        self.fig = fig
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe_)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1)
        self.draw_all_parames(fig, x_array, nrows)
        self.after(300, self.set_better_geometry)

    def set_better_geometry(self):
        geometry = self.geometry()
        geometry = re.sub(r"(\d+x)(\d+)(.+)", lambda m: m.group(1) + "400" + m.group(3), geometry)
        self.geometry(geometry)

    def draw_all_parames(self, fig, x_array, nrows):
        x_ = self.x_
        xmin, xmax = self.xlim

        names = self.optimizer.params_type.get_parameter_names()

        c = self.curr_index
        j = self.best_index
        k = len(x_) - 1
        ncols = 8
        gs = GridSpec(nrows, ncols)
        axes = []
        for i in range(nrows):
            ax1 = fig.add_subplot(gs[i,0])
            ax1.set_axis_off()
            ax1.text(0.5, 0.5, names[i], ha="center", va="center")
            ax2 = fig.add_subplot(gs[i,1:])
            ax2.set_xticklabels([])
            ax2.plot(x_, x_array[:,i])
            ax2.set_xlim(xmin, xmax)
            ymin, ymax = ax2.get_ylim()
            ax2.set_ylim(ymin, ymax)
            ax2.plot(x_[[j,j]], [ymin, ymax], color="red")
            if c != j:
                ax2.plot(x_[[c,c]], [ymin, ymax], color="yellow")
            if j != k:
                ax2.plot(x_[[k,k]], [ymin, ymax], color="gray", alpha=0.3)
            axes.append((ax1, ax2))
        self.axes = np.array(axes)

        fig.tight_layout()
        fig.subplots_adjust(top=0.99, bottom=0.01, hspace=0.3, right=0.96)
        self.mpl_canvas.draw()



import tkinter as Tk

def test_it():

    root = Tk.Tk()

    for k, c in enumerate(["A", "B", "C", "D", "E", "F"]):
        label = Tk.Label(root, text=c)
        label.grid(row=0, column=k)

    for k, c in enumerate(["X", "Y", "Z"]):
        label = Tk.Label(root, text=c)
        label.grid(row=1, column=k*2, columnspan=2)

    for k, c in enumerate(["A", "B", "C", "D", "E", "F"]):
        label = Tk.Label(root, text=c)
        label.grid(row=2, column=k)

    root.mainloop()

if __name__ == "__main__":
    test_it()


# coding: utf-8
"""
    IndividualSubtractorDialog.py

    Copyright (c) 2019-2021, SAXS Team, KEK-PF
"""
import os
from difflib import SequenceMatcher
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk, is_empty_val
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry
try:
    import molass_legacy.KekLib.CustomMessageBox         as MessageBox
except:
    import OurMessageBox            as MessageBox
from SerialDataUtils import get_xray_files
from DataSubtractor import subtract_impl

class IndividualSubtractorDialog(Dialog):
    def __init__(self, parent):
        self.busy = False
        Dialog.__init__(self, parent, "Individual Subtractor", visible=False)

    def body(self, body_frame):
        tk_set_icon_portable(self)

        iframe = Tk.Frame(body_frame)
        iframe.pack(padx=20, pady=10)

        folder_entry_width = 60
        num_files_label_width = 6

        grid_row = -1

        grid_row += 1
        label = Tk.Label(iframe, text="Folder Path")
        label.grid(row=grid_row, column=1)
        label = Tk.Label(iframe, text="File Name Pattern")
        label.grid(row=grid_row, column=2, padx=10)
        label = Tk.Label(iframe, text="Files")
        label.grid(row=grid_row, column=3)


        grid_row += 1
        label = Tk.Label(iframe, text="Sample Data")
        label.grid(row=grid_row, column=0)

        self.in_folder1 = Tk.StringVar()
        self.in_files1 = []
        self.name_pattern1 = Tk.StringVar()
        self.num_files1 = Tk.IntVar()
        self.in_folder1_cb = lambda: self.on_entry_in_folder(self.in_folder1, self.in_files1, self.name_pattern1, self.num_files1)
        in_folder1_entry = FolderEntry(iframe, textvariable=self.in_folder1, width=folder_entry_width,
                                on_entry_cb=self.in_folder1_cb)
        in_folder1_entry.grid(row=grid_row, column=1)
        in_folder1_entry.bind('<Leave>', lambda *args:self.in_folder1_cb())

        label = Tk.Label(iframe, textvariable=self.name_pattern1)
        label.grid(row=grid_row, column=2, padx=10, sticky=Tk.W)

        num_files_label = Tk.Label(iframe, textvariable=self.num_files1, width=num_files_label_width, justify=Tk.RIGHT)
        num_files_label.grid(row=grid_row, column=3)

        grid_row += 1
        label = Tk.Label(iframe, text="Buffer Data")
        label.grid(row=grid_row, column=0)
        self.in_folder2 = Tk.StringVar()
        self.in_files2 = []
        self.name_pattern2 = Tk.StringVar()
        self.num_files2 = Tk.IntVar()
        self.in_folder2_cb = lambda: self.on_entry_in_folder(self.in_folder2, self.in_files2, self.name_pattern2, self.num_files2)
        in_folder2_entry = FolderEntry(iframe, textvariable=self.in_folder2, width=folder_entry_width,
                                on_entry_cb=self.in_folder2_cb)
        in_folder2_entry.grid(row=grid_row, column=1, pady=5)
        in_folder2_entry.bind('<Leave>', lambda *args:self.in_folder2_cb())

        label = Tk.Label(iframe, textvariable=self.name_pattern2)
        label.grid(row=grid_row, column=2, padx=10, sticky=Tk.W)

        num_files_label = Tk.Label(iframe, textvariable=self.num_files2, width=num_files_label_width, justify=Tk.RIGHT)
        num_files_label.grid(row=grid_row, column=3)

        grid_row += 1
        label = Tk.Label(iframe, text="Subtracted Data")
        label.grid(row=grid_row, column=0)
        self.out_folder = Tk.StringVar()
        out_folder_entry = FolderEntry(iframe, textvariable=self.out_folder, width=folder_entry_width,
                                on_entry_cb=self.on_entry_out_folder)
        out_folder_entry.grid(row=grid_row, column=1)
        out_folder_entry.bind('<Leave>', lambda *args: self.on_entry_out_folder())

        self.name_pattern3 = Tk.StringVar()
        entry = Tk.Entry(iframe, textvariable=self.name_pattern3, width=20)
        entry.grid(row=grid_row, column=2, padx=10, sticky=Tk.W)

        self.num_files3 = Tk.IntVar()
        num_files_label = Tk.Label(iframe, textvariable=self.num_files3, width=num_files_label_width, justify=Tk.RIGHT)
        num_files_label.grid(row=grid_row, column=3)

        grid_row += 1
        space = Tk.Label(iframe)
        space.grid(row=grid_row, column=0)

        grid_row += 1
        label = Tk.Label(iframe, text="Progress")
        label.grid(row=grid_row, column=0)
        self.mpb = ttk.Progressbar(iframe,orient ="horizontal", length=600, mode="determinate")
        self.mpb.grid(row=grid_row, column=1, columnspan=3, padx=5)

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        self.subtract_btn = Tk.Button(box, text="Subtract", width=10, command=self.subtract, state=Tk.DISABLED)
        self.subtract_btn.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def show(self):
        self._show()

    def on_entry_in_folder(self, in_folder, in_files, name_pattern, num_files):
        files = get_xray_files(in_folder.get())
        in_files.clear()
        in_files += files
        if len(files) > 0:
            dir_, file = os.path.split(files[0])
            name_pattern.set(file)
        num_files.set( len(files) )

    def on_entry_out_folder(self):
        _, in_folder1 = os.path.split(self.in_folder1.get())
        name_pattern1 = self.name_pattern1.get()
        _, in_folder2 = os.path.split(self.in_folder2.get())
        name_pattern2 = self.name_pattern2.get()
        # print((in_folder1, name_pattern1), (in_folder2, name_pattern2))
        _, out_folder = os.path.split(self.out_folder.get())

        if is_empty_val(in_folder1) or is_empty_val(in_folder2) or is_empty_val(out_folder):
            return

        ret_pattern = self.generate_pattern(in_folder1, name_pattern1, in_folder2, name_pattern2, out_folder)
        self.name_pattern3.set(ret_pattern)
        self.subtract_btn.config(state=Tk.NORMAL)

    def subtract(self):
        num_files1 = len(self.in_files1)
        num_files2 = len(self.in_files2)

        if num_files1 != num_files2:
            MessageBox.showerror("", "Numbers of files are different: %d != %d" % (num_files1, num_files2))
            return

        name_pattern = self.name_pattern3.get().replace('_00000', '_%05d')
        print('name_pattern=', name_pattern)

        self.busy = True
        process_option = 0
        subtract_impl(self.in_files1, self.in_files2, process_option, self.out_folder.get(), name_pattern, dialog=self)
        self.busy = False

    def is_busy(self):
        return self.busy

    def generate_pattern(self, in_folder1, name_pattern1, in_folder2, name_pattern2, out_folder):
        debug = False
        match = SequenceMatcher(None, in_folder1, in_folder2).find_longest_match(0, len(in_folder1), 0, len(in_folder2))
        if debug:
            print(match)
            print(in_folder1[match.a: match.a + match.size])

        match = SequenceMatcher(None, name_pattern1, name_pattern2).find_longest_match(0, len(name_pattern1), 0, len(name_pattern2))
        common_pattern = name_pattern1[match.a: match.a + match.size]
        if debug:
            print(match)
            print(common_pattern)

        return "Diff" + common_pattern



import numpy as np
import matplotlib.pyplot as plt
from utilities import *
from paircorrelation import pairCorrelationFunction_3D

# Particle setup
domain_size = 20.0
num_particles = 10000

# Calculation setup
dr = 0.1

### Random arrangement of particles ###
particle_radius = 0.1
rMax = domain_size / 4
x = np.random.uniform(low=0, high=domain_size, size=num_particles)
y = np.random.uniform(low=0, high=domain_size, size=num_particles)
z = np.random.uniform(low=0, high=domain_size, size=num_particles)

# Compute pair correlation
g_r, r, reference_indices = pairCorrelationFunction_3D(x, y, z, domain_size, rMax, dr)

# Visualize
plt.figure()
plt.plot(r, g_r, color='black')
plt.xlabel('r')
plt.ylabel('g(r)')
plt.xlim( (0, rMax) )
plt.ylim( (0, 1.05 * g_r.max()) )
plt.show()



"""
    V2PropOptimizer.PropOptimizer.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
from time import sleep
import numpy as np
from scipy import stats
from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier
from molass_legacy.GuinierAnalyzer.SimpleGuinierScore import compute_rg

PROP_MIN_RATIO = 0.1
PROP_MIN_VALUE = 1e-5

def compute_range_rgs(qv, D, E, paired_ranges, C, return_sgs=False, region_info=None):
    assert C.shape[0] == 2
    assert len(paired_ranges) == 1
    cmax = np.max(C, axis=1)
    cprop = cmax/np.sum(cmax)
    i = np.argmin(cprop)        # minor component

    if return_sgs:
        ret_sgs = []
    else:
        ret_rgs = []

    for f, t in paired_ranges[0].get_fromto_list():
        slice_ = slice(f, t+1)
        D_ = D[:,slice_]
        E_ = E[:,slice_]
        C_ = C[:,slice_]
        cmax = np.max(C_, axis=1)
        cprop = cmax/np.sum(cmax)
        # print("cmax=", cmax)
        # print("cprop=", cprop)
        if cprop[i] < 1e-3:
            C_ = C_[1,:].reshape((1, C_.shape[1]))
            j = 0
        else:
            j = 1
        Cinv = np.linalg.pinv(C_)
        P_ = D_ @ Cinv
        Minv = np.linalg.pinv(D_)
        W = np.dot(Minv, P_)
        Pe  = np.sqrt(E_**2 @ W**2)
        data = np.array([qv, P_[:,j], Pe[:,j]]).T

        if region_info is None:
            sg = SimpleGuinier(data)
            if return_sgs:
                ret_sgs.append(sg)
            else:
                ret_rgs.append(sg.Rg)
        else:
            assert not return_sgs

            region, qv2 = region_info
            x_ = qv2[region]
            y_ = np.log(P_[region,j])
            slope = stats.linregress(x_, y_)[0]
            rg = compute_rg(slope)
            ret_rgs.append(rg)

    if return_sgs:
        return ret_sgs
    else:
        return ret_rgs

class RangeRgComputer:
    def __init__(self, qv, D, E, paired_ranges):
        self.return_sgs = True
        self.region_list = []
        self.fixed_params = (qv, D, E, paired_ranges)
        self.qv2 = qv**2

    def compute(self, C, debug=False):
        if self.return_sgs:
            ret_sgs = compute_range_rgs(*self.fixed_params, C, return_sgs=True)
            ret_rgs = [sg.Rg for sg in ret_sgs]
            regions = [(sg.guinier_start, sg.guinier_stop) for sg in ret_sgs]
            self.region_list.append(regions)
            if len(self.region_list) == 20:
                region_array = np.array(self.region_list)
                start = int(np.mean(region_array[:,1,0]))
                stop = int(np.mean(region_array[:,1,1]))
                if debug:
                    import molass_legacy.KekLib.DebugPlot as plt
                    with plt.Dp():
                        fig, ax = plt.subplots()
                        ax.set_title("region_list debug")
                        region0 = region_array[:,0,:]
                        region1 = region_array[:,1,:]
                        ax.plot(*region0.T, "o")
                        ax.plot(*region1.T, "o")
                        ax.plot(start, stop, "o", color="red")
                        fig.tight_layout()
                        plt.show()
                self.return_sgs = False
                self.region = slice(start, stop)
        else:
            ret_rgs = compute_range_rgs(*self.fixed_params, C, region_info=[self.region, self.qv2])
        return ret_rgs



# coding: utf-8
"""
    SaxsSamples.py

    Copyright (c) 2019-2021, SAXS Team, KEK-PF
"""
import os
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # noqa: F401 unused import

def get_plot_axis():
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    return ax

def midpoints(x):
    sl = ()
    for i in range(x.ndim):
        x = (x[sl + np.index_exp[:-1]] + x[sl + np.index_exp[1:]]) / 2.0
        sl += np.index_exp[:]
    return x

class ObjectVoxels:
    def __init__(self, shape):
        self.shape = shape_ = np.array(shape)
        n = np.max(shape_)/2
        x, y, z = np.indices(shape_+1)/n - 1
        self.xyz = (x, y, z)
        mx = midpoints(x)
        my = midpoints(y)
        mz = midpoints(z)
        self.points = (mx, my, mz)

    def get_xyz(self):
        return self.xyz

    def get_data(self, density=1):
        data = np.zeros(self.shape)
        data[self.bool_array] = density
        return data

class BallVoxels(ObjectVoxels):
    def __init__(self, shape=(64,64,64), center=(0,0,0), radius=1):
        ObjectVoxels.__init__(self, shape)
        mx, my, mz = self.points
        cx, cy, cz = center
        self.bool_array = (mx-cx)**2 + (my-cy)**2 + (mz-cz)**2 < radius**2

class EllipsoidVoxels(ObjectVoxels):
    def __init__(self, shape=(64,64,64), center=(0,0,0), radii=(1,1,1)):
        ObjectVoxels.__init__(self, shape)
        mx, my, mz = self.points
        cx, cy, cz = center
        a, b, c = radii
        self.bool_array = ((mx-cx)/a)**2 + ((my-cy)/b)**2 + ((mz-cz)/c)**2 < 1

class DiscVoxels(ObjectVoxels):
    def __init__(self, shape=(64,64,64), center=(0,0,0), radius=1, height=0.05):
        ObjectVoxels.__init__(self, shape)
        mx, my, mz = self.points
        cx, cy, cz = center
        in_circle = (mx-cx)**2 + (my-cy)**2 < radius**2
        in_depth = (mz-cz)**2 < height**2
        self.bool_array = np.logical_and(in_circle, in_depth)

class TorusVoxels(ObjectVoxels):
    def __init__(self, shape=(64,64,64), center=(0,0,0), R=0.8, r=0.2):
        ObjectVoxels.__init__(self, shape)
        mx, my, mz = self.points
        cx, cy, cz = center
        self.bool_array = (np.sqrt((mx-cx)**2 + (my-cy)**2) - R)**2 + (mz-cz)**2 < r**2



"""
    DmaxEstimation.py

    Copyright (c) 2020-2025, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from .denss.core import Sasrec, clean_up_data, calc_rg_I0_by_guinier, calc_rg_by_guinier_peak, filter_P
from .DenssUtils import fit_data_impl

### DENSS.denss.core.py copy & modify BEGIN ###
def estimate_dmax(Iq,dmax=None,clean_up=True):
    """Attempt to roughly estimate Dmax directly from data."""
    # first, clean up the data
    if clean_up:
        Iq = clean_up_data(Iq)
    q = Iq[:, 0]
    I = Iq[:, 1]
    nq = len(q)
    if dmax is None:
        # first, estimate a very rough rg from the first 20 data points
        nmax = 20
        try:
            rg, I0 = calc_rg_I0_by_guinier(Iq, ne=nmax)
        except:
            rg = calc_rg_by_guinier_peak(Iq, exp=1, ne=100)
        # next, dmax is roughly 3.5*rg for most particles
        # so calculate P(r) using a larger dmax, say twice as large, so 7*rg
        D = 7 * rg
    else:
        # allow user to give an initial estimate of Dmax
        # multiply by 2 to allow for enough large r values
        D = 2 * dmax
    # create a calculated q range for Sasrec for low q out to q=0
    qmin = np.min(q)
    dq = (q.max() - q.min()) / (q.size - 1)
    nqc = int(qmin / dq)
    qc = np.concatenate(([0.0], np.arange(nqc) * dq + (qmin - nqc * dq), q))
    # run Sasrec to perform IFT
    sasrec = Sasrec(Iq[:nq // 2], D, qc=None, alpha=0.0, extrapolate=False)
    # if the rg estimate was way off, it would screw up Dmax estimate
    # but the sasrec rg should be more accurate, even with a screwed up guinier estimate
    # so run it again, but this time with the Dmax = 7*sasrec.rg
    # only do this if rg is significantly different
    if np.abs(sasrec.rg - rg) > 0.2 * sasrec.rg:
        sasrec = Sasrec(Iq[:nq // 2], D=7 * sasrec.rg, qc=None, alpha=0.0, extrapolate=False)
    # lets test a bunch of different dmax's on a logarithmic spacing
    # then see where chi2 is minimal. that at least gives us a good ball park of Dmax
    # the main problem is that we don't know the scale even remotely, or the units,
    # so we need to check many orders of magnitude
    Ds = np.logspace(.1, np.log10(2 * 7 * rg), 10)
    chi2 = np.zeros(len(Ds))
    for i in range(len(Ds)):
        sasrec = Sasrec(Iq[:nq // 2], D=Ds[i], qc=None, alpha=0.0, extrapolate=False)
        chi2[i] = sasrec.calc_chi2()
    order = np.argsort(chi2)
    D = 2 * np.interp(2 * chi2.min(), chi2[order], Ds[order])
    # one final time with new D and full q range
    sasrec = Sasrec(Iq, D=D, qc=None, alpha=0.0, extrapolate=False)
    # now filter the P(r) curve for estimating Dmax better
    qmax = 2 * np.pi / D
    # qmax_fraction = 0.5
    r, Pfilt, sigrfilt = filter_P(sasrec.r, sasrec.P, sasrec.Perr, qmax=qmax)  # qmax_fraction*Iq[:,0].max())
    # import matplotlib.pyplot as plt
    # plt.plot(sasrec.r,sasrec.r*0,'k--')
    # plt.plot(sasrec.r, sasrec.P,'b-')
    # plt.plot(r,Pfilt,'r-')
    # estimate D as the first position where P becomes less than 0.01*P.max(), after P.max()
    Pargmax = Pfilt.argmax()
    # catch cases where the P(r) plot goes largely negative at large r values,
    # as this indicates repulsion. Set the new Pargmax, which is really just an
    # identifier for where to begin searching for Dmax, to be any P value whose
    # absolute value is greater than at least 10% of Pfilt.max. The large 10% is to
    # avoid issues with oscillations in P(r).
    argmax_threshold = 0.05
    above_idx = np.where((np.abs(Pfilt) > argmax_threshold * Pfilt.max()) & (r > r[Pargmax]))
    Pargmax = np.max(above_idx)
    dmax_threshold = (0.01 * Pfilt.max())
    near_zero_idx = np.where((np.abs(Pfilt[Pargmax:]) < dmax_threshold))[0]
    near_zero_idx += Pargmax
    D_idx = near_zero_idx[0]
    D = r[D_idx]
    sasrec.D = np.copy(D)
    # plt.plot(sasrec.r,sasrec.r*0+argmax_threshold*Pfilt.max(),'g--')
    # plt.plot(sasrec.r,sasrec.r*0-argmax_threshold*Pfilt.max(),'g--')
    # plt.plot(sasrec.r,sasrec.r*0+dmax_threshold,'r--')
    # plt.axvline(D,c='r')
    # plt.plot()
    # plt.show()
    # exit()
    sasrec.update()
    # return D, sasrec
    return D, sasrec, [r, Pfilt, Pargmax, D_idx]
### DENSS.denss.core.py copy & modify END ###

def plot_input(ax, data, in_file):
    q = data[:,0]
    a = data[:,1]
    e = data[:,2]

    sasrec, work_info = fit_data_impl(q, a, e, in_file)
    qc = sasrec.qc
    ac = sasrec.Ic
    ec = work_info.Icerr

    ax.set_yscale('log')
    ax.set_xlabel('q', fontsize=16)
    ax.set_ylabel('log(I)', fontsize=16)

    ax.plot(q, a, color='C1', label="input data")
    # ax1.plot(qc, ac, color='C2', label="fitted data")

    ax.legend(fontsize=16)

def illustrate_dmax(ax, data):
    D_, sasrec_, info = estimate_dmax(data)
    r = sasrec_.r
    P = sasrec_.P
    r_, Pfilt, Pargmax, D_idx = info

    ax.set_xlabel('r', fontsize=16)
    ax.set_ylabel('P', fontsize=16)

    ax.plot(r, P, color='C1', label="P(r) from input data")
    ax.plot(r_, Pfilt, color='C2', label="filtered P(r)")
    ax.plot(r_[D_idx], Pfilt[D_idx], 'o', color='red', label='estimated Dmax')

    ax.legend(fontsize=16)
    ymin, ymax = ax.get_ylim()
    hymax = ymax*0.5
    if ymin < 0 and abs(ymin) > hymax:
        # cut off the negative part of P(r) if it is too large
        ymin = -abs(hymax)
        ax.set_ylim(ymin, ymax)

def demo(in_file):
    print(in_file)
    data = np.loadtxt(in_file)

    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(16,7))
    fig.suptitle("Denss Dmax Estimation Illustrated", fontsize=30)

    plot_input(ax1, data, in_file)
    illustrate_dmax(ax2, data)

    fig.tight_layout()
    fig.subplots_adjust(top=0.9)
    plt.show()



"""
    Models.Stochastic.LnporeUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""

def plot_lognormal_fitting_state(x, y, lnp_params, rgs, plot_boundary=False, title=None, save_fig_as=None):
    from matplotlib.gridspec import GridSpec
    import molass_legacy.KekLib.DebugPlot as plt
    from importlib import reload
    import Models.Stochastic.LnporeUtilsImpl as module
    reload(module)
    from molass_legacy.Models.Stochastic.LnporeUtilsImpl import plot_elution_exclusion_impl, plot_lognormal_psd_impl

    def plot_lnpore_component_with_sliders():
        reload(module)
        from molass_legacy.Models.Stochastic.LnporeUtilsImpl import plot_lnpore_component_with_sliders_impl
        plot_lnpore_component_with_sliders_impl(x, y, lnp_params, rgs)

    extra_button_specs = [("Plot Component with Sliders", plot_lnpore_component_with_sliders)]

    with plt.Dp(button_spec=["OK", "Cancel"],
                extra_button_specs=extra_button_specs):
        if plot_boundary:
            fig = plt.figure(figsize=(20,5))
            gs = GridSpec(1,7)
            ax1 = fig.add_subplot(gs[0,0:3])
            ax2 = fig.add_subplot(gs[0,3:5])
            ax3 = fig.add_subplot(gs[0,5:7])
        else:
            fig = plt.figure(figsize=(14,5))
            gs = GridSpec(1,5)
            ax1 = fig.add_subplot(gs[0,0:3])
            ax2 = fig.add_subplot(gs[0,3:5])
        if title is None:
            title = "Lognormal Fitting"
        fig.suptitle(title, fontsize=20)
        plot_elution_exclusion_impl(ax1, x, y, lnp_params, rgs)
        mu, sigma = lnp_params[5:7]
        plot_lognormal_psd_impl(ax2, mu, sigma)
        if plot_boundary:
            from Simulative.ApproxBoundary import compute_approx_boundary, plot_boundary_impl
            rv, bv = compute_approx_boundary(x, y, lnp_params, rgs, quickly=True)
            plot_boundary_impl(ax3, rv, bv, mu, sigma)

        fig.tight_layout()
        if save_fig_as is None:
            ret = plt.show()
        else:
            plt.show(block=False)
            fig.savefig(save_fig_as)
            ret = True            
    return ret


# coding: utf-8
"""
    LightObjects.py

    Objects to facilitate garbage collection

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF
"""
class AnyObject:
    def __init__( self, **entries ): 
        self.__dict__.update(entries)

class LightIntensity:
    def __init__( self, intensity ):
        if intensity is None:
            self.orig_array = None
        else:
            self.orig_array = intensity.orig_array
            self.comments   = '' 

class LightFit:
    def __init__( self, fit ):
        self.Rg     = fit.Rg
        self.I0     = fit.I0
        self.degree = fit.degree
        self.result = AnyObject( aic=fit.result.aic, bic=fit.result.bic )

class LightResult:
    def __init__( self, result ):
        self.Rg             = result.Rg
        self.Rg_            = result.Rg_
        self.Rg_stdev       = result.Rg_stdev
        self.I0             = result.I0
        self.I0_            = result.I0_
        self.I0_stdev       = result.I0_stdev
        self.From           = result.From
        self.To             = result.To
        self.min_qRg        = result.min_qRg
        self.max_qRg        = result.max_qRg
        self.fit            = LightFit( result.fit )
        self.bico_mono_ratio    = result.bico_mono_ratio
        self.IpI            = result.IpI
        self.bicomponent    = result.bicomponent
        self.head_trend     = result.head_trend
        self.result_type    = result.result_type
        self.quality_object = result.quality_object
        self.Quality        = result.Quality



"""
    Models.Stochastic.ExclusionCurvatures.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.integrate import quad
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, Button

def tripore_simulation():
    from molass_legacy.Models.Stochastic.LognormalPoreFunc import distr_func
    from molass_legacy.Models.Stochastic.ParamLimits import MAX_PORESIZE
    from molass_legacy.Models.Stochastic.LognormalUtils import compute_mu_sigma
    N = 1000
    T = 1
    t0 = 50
    me = 1.5
    mp = 1.5
    mnp_sizes = np.array([150, 150, 150])
    mnp_props = np.array([1, 1, 1])
    mnp_props = mnp_props/np.sum(mnp_props)
    sizes = np.array([100, 140, 200])
    props = np.array([2, 1, 3])
    props = props/np.sum(props)

    rv = np.linspace(10, 120, 100)
    def compute_tripore_tv(N, T, t0, me, mp, sizes, props):
        tv = []
        for rg in rv:
            rhov = rg/sizes
            rhov[rhov > 1] = 1
            tr = t0 + N * T * np.sum(props * (1 - rhov)**(me + mp))
            tv.append(tr)
        return np.asarray(tv)

    def compute_lnpore_tv(N, T, t0, me, mp, mu, sigma):
        tv = []
        for rg in rv:
            rhov = rg/sizes
            rhov[rhov > 1] = 1
            tr = t0 + N * T * quad(lambda r : distr_func(r, mu, sigma) * (1 - min(1, rg/r))**(me +   mp), rg, MAX_PORESIZE)[0]
            tv.append(tr)
        return np.asarray(tv)

    tv_mnp = compute_tripore_tv(N, T, t0, me, mp, mnp_sizes, mnp_props)
    tv = compute_tripore_tv(N, T, t0, me, mp, sizes, props)
    mu, sigma = compute_mu_sigma(150, 10)
    tv_lnp = compute_lnpore_tv(N, T, t0, me, mp, mu, sigma)

    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(18,5))
    fig.suptitle("Exclusion Curve Simulations of Monopore, Tripore, and Lognormalpore PSDs", fontsize=20)
    ax1.set_title("Exclusion Curves", fontsize=16)
    mnp_curve, = ax1.plot(tv_mnp, rv, alpha=0.5, label="monopore")
    trp_curve, = ax1.plot(tv, rv, alpha=0.5, label="tripore")
    lnp_curve, = ax1.plot(tv_lnp, rv, alpha=0.5, label="lognormalpore")
    ax1.axvline(x=t0, color="red", label="$t_0$")
    ax1.legend()
    ax1.set_xlabel("Retension Time (Frames)")
    ax1.set_ylabel(r"Rg ($\AA$)")

    ax2.set_title("Pore Size Distributions", fontsize=16)
    ax2.bar(mnp_sizes, mnp_props, width=5, alpha=0.5, label="monopore")
    ax2.bar(sizes, props, width=5, alpha=0.5, label="tripore")
    pv = np.linspace(10, 400, 100)   
    ax2.fill_between(pv, np.zeros(len(pv)), distr_func(pv, mu, sigma), alpha=0.5, label="lognormalpore")
    ax2.legend()
    ax2.set_xlabel(r"Pore Size ($\AA$)")
    ax2.set_ylabel("Density")

    # https://stackoverflow.com/questions/22263807/how-is-order-of-items-in-matplotlib-legend-determined
    handles, labels = ax2.get_legend_handles_labels()
    order = [1, 2, 0]
    ax2.legend([handles[idx] for idx in order],[labels[idx] for idx in order])
    ax2.set_xlim(0, 400)
    ax2.set_ylim(0, 0.6)

    patch_list = []
    for c in ax2.containers:
        print(len(c.patches))
        patch_list.append(c.patches)
    print(len(patch_list))

    plot_params = np.concatenate([[N, T, t0, me, mp, 150], sizes, props, [150, 10]])

    slider_specs = [    ("N", 0, 6000, plot_params[0]),
                        ("T", 0, 3, plot_params[1]),
                        ("t0",  -1000, 1000, plot_params[2]),
                        ("me", 0, 3, plot_params[3]),
                        ("mp", 0, 3, plot_params[4]),
                        ("mono poresize", 20, 400, plot_params[5]),
                        ("tri poresize1", 20, 400, plot_params[6]),
                        ("tri poresize2", 20, 400, plot_params[7]),
                        ("tri poresize3", 20, 400, plot_params[8]),                        
                        ("tri proportion1", 0, 1, plot_params[9]),
                        ("tri proportion2", 0, 1, plot_params[10]),
                        ("tri proportion3", 0, 1, plot_params[11]),       
                        ("lognormal mode", 20, 400, plot_params[12]),
                        ("lognormal stdev", 1, 100, plot_params[13]),       
                        ]

    def slider_update(k, val):
        print([k], "slider_update", val)
        plot_params[k] = val
        N, T, t0, me, mp = plot_params[0:5]

        if k == 5:
            i = 0
            for patch in patch_list[i]:
                patch.set_x(val)
            mnp_sizes = np.ones(3)*val
            tv_mnp = compute_tripore_tv(N, T, t0, me, mp, mnp_sizes, mnp_props)
            mnp_curve.set_xdata(tv_mnp)

        if 6 <= k and k <= 11:
            i = 1
            if k < 9:
                j = k - 6
                patch = patch_list[i][j]
                patch.set_x(val)
            elif k < 12:
                j = k - 9
                patch = patch_list[i][j]
                patch.set_height(val)

                indeces = np.setdiff1d(np.arange(3), j)
                props = plot_params[9 + indeces]   
                props /= np.sum(props)       
                print("indeces", indeces, "props=", props)
                for m, p in zip(indeces, props):
                    v = p*(1-val)
                    patch_list[i][m].set_height(v)
                    k_ = 9 + m
                    plot_params[k_] = v
                    slider = sliders[k_]
                    slider.eventson = False
                    slider.set_val(v)
                    slider.eventson = True
            else:
                # to be implemented
                pass

            sizes = plot_params[6:9]
            props = plot_params[9:12]
            tv = compute_tripore_tv(N, T, t0, me, mp, sizes, props)
            trp_curve.set_xdata(tv)

        fig.canvas.draw_idle()

    slider_axes = []
    sliders = []
    for k, (label, valmin, valmax, valinit) in enumerate(slider_specs, start=0):
        ax_ = fig.add_axes([0.75, 0.8 - 0.05*k, 0.2, 0.03])
        slider  = Slider(ax_, label=label, valmin=valmin, valmax=valmax, valinit=valinit)
        slider.on_changed(lambda val, k_=k: slider_update(k_, val))
        slider_axes.append(ax_)
        sliders.append(slider)

    def reset(event):
        print("reset")
        for k, slider in enumerate(sliders):
            slider.reset()

    button_ax = fig.add_axes([0.85, 0.07, 0.12, 0.05])
    debug_btn = Button(button_ax, 'Reset', hovercolor='0.975')
    debug_btn.on_clicked(reset)

    fig.tight_layout()
    fig.subplots_adjust(left=0.06, right=0.67, wspace=0.2)
    plt.show()

if __name__ == "__main__":
    import sys
    import seaborn as sns
    sns.set_theme()
    sys.path.append("../lib")

    tripore_simulation()


# coding: utf-8
"""

    ファイル名：   MappingAutoHelper.py

    処理内容：

        濃度マッピング不具合の自動修正

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF

"""
from molass_legacy._MOLASS.SerialSettings     import get_setting
# from molass_legacy.SerialAnalyzer.ElutionCurve       import PEAK_INFO_RAW

PEAK_MATCH_ALLOW_RATIO  = 0.05

def get_add_info( name_t, x_t, x_f, scale, allow_t ):
    ret_info    = []

    debug = True
    if debug:
        print( '---- get_add_info for', name_t, scale, allow_t )
        print( 'x_t=', x_t )
        print( 'x_f=', x_f )

    for pf in x_f:
        pf_ = pf * scale
        found = False
        for pt in x_t:
            if abs( pt - pf_ ) < allow_t:
                found = True
        if not found:
            ret_info.append( pf_ )

    if debug: print( 'ret_info=', ret_info )

    return ret_info

def make_helper_info( mapper, exception ):
    if get_setting( 'enable_auto_helper' ) == 0:
        return None

    exc_info = exception.args[1]
    if exc_info[0] > 2:
        # can't cope with this situation
        return None

    print( '------------------------- make_helper_info: ', exc_info )
    a_info, x_info  = mapper.get_original_curve_info()

    a_size  = len(mapper.a_vector)
    x_size  = len(mapper.x_vector)
    a_allow = a_size * PEAK_MATCH_ALLOW_RATIO
    x_allow = x_size * PEAK_MATCH_ALLOW_RATIO

    ratio   = a_size / x_size
    print( 'ratio=', ratio )

    A       = ratio
    A_inv   = 1/A

    a_peak_x        = [ info[1] for info in a_info.peak_info ]
    x_peak_x        = [ info[1] for info in x_info.peak_info ]
    a_peak_x_raw    = [ info[1] for info in a_info.raw_info[PEAK_INFO_RAW] ]
    x_peak_x_raw    = [ info[1] for info in x_info.raw_info[PEAK_INFO_RAW] ]

    a_add_info      = get_add_info( 'a_peak_x',     a_peak_x,       x_peak_x,       A, a_allow )
    a_add_info_raw  = get_add_info( 'a_peak_x_raw', a_peak_x_raw,   x_peak_x_raw,   A, a_allow )

    x_add_info      = get_add_info( 'x_peak_x',     x_peak_x,       a_peak_x,       A_inv, x_allow )
    x_add_info_raw  = get_add_info( 'x_peak_x_raw', x_peak_x_raw,   a_peak_x_raw,   A_inv, x_allow )

    anyway_flag = False
    helper_info = [ ( [], a_add_info, a_add_info_raw ), ( [], x_add_info, x_add_info_raw ), mapper.flow_changes, anyway_flag ]

    num_info    = len(a_add_info) + len(a_add_info_raw) + len(x_add_info) + len(x_add_info_raw)
    if num_info == 0:
        helper_info = None
    else:
        mapper.logger.info( 'succeeded in making helper info' )

    return helper_info



# coding: utf-8
"""
    ReciprocalData.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import os
from bisect import bisect_right
import numpy as np
from scipy import ndimage, interpolate
import matplotlib.pyplot as plt
import matplotlib.cm as cm
from molass_legacy.KekLib.NumpyUtils import np_loadtxt_robust

class ReciprocalData:
    def __init__(self, shape):
        dmax = 100.0
        voxel = 5
        oversampling = 3

        D = dmax
        dn = shape[0]

        ############### from denss begin ###################

        #Initialize variables

        side = oversampling*D
        halfside = side/2

        if dn is None:
            dn = int(side/voxel)
            #want dn to be even for speed/memory optimization with the FFT, ideally a power of 2, but wont enforce that
            if dn%2==1:
                dn += 1

        #store dn for later use if needed
        nbox = dn

        dx = side/dn
        dV = dx**3
        V = side**3
        x_ = np.linspace(-halfside,halfside,dn)
        x,y,z = np.meshgrid(x_,x_,x_,indexing='ij')
        r = np.sqrt(x**2 + y**2 + z**2)

        df = 1/side
        qx_ = np.fft.fftfreq(x_.size)*dn*df*2*np.pi

        qx, qy, qz = np.meshgrid(qx_,qx_,qx_,indexing='ij')
        qr = np.sqrt(qx**2+qy**2+qz**2)
        qmax = np.max(qr)
        qstep = np.min(qr[qr>0])
        nbins = int(qmax/qstep)
        qbins = np.linspace(0,nbins*qstep,nbins+1)

        #create modified qbins and put qbins in center of bin rather than at left edge of bin.
        self.qbinsc = np.copy(qbins)
        self.qbinsc[1:] += qstep/2.

        #create an array labeling each voxel according to which qbin it belongs
        self.qbin_labels = np.searchsorted(qbins,qr,"right")
        self.qbin_labels -= 1

    def get_reciprocal(self, data):
        return np.fft.fftn(data)

    def get_scattering_curve(self, q, F, absolute=False):
        qbinsc = self.qbinsc
        qbin_labels = self.qbin_labels

        #APPLY RECIPROCAL SPACE RESTRAINTS
        #calculate spherical average of intensities from 3D Fs
        if absolute:
            I3D = F**2
        else:
            I3D = np.abs(F)**2
        print('I3D.shape=', I3D.shape, 'qbin_labels.shape=', qbin_labels.shape)
        index = np.arange(0,qbin_labels.max()+1)
        # print('index=', index)
        Imean = ndimage.mean(I3D, labels=qbin_labels, index=index)
        self.Imean = Imean

        #scale Fs to match data
        self.interp = interpolate.interp1d(qbinsc, Imean, kind='cubic', fill_value="extrapolate")
        I4chi = self.interp(q)

        ############### from denss end ###################

        curve_y = I4chi

        if False:
            import molass_legacy.KekLib.DebugPlot as dplt
            dplt.push()
            fig = dplt.figure()
            ax = fig.gca()
            ax.plot(curve_y)
            dplt.show()
            dplt.pop()

        return curve_y

    def get_3d_magnitudes(self, qbin_values, convolve=False):
        sqv = np.sqrt(qbin_values)
        ret = sqv[self.qbin_labels]
        if convolve:
            dn = 256
            nm = dn - 8
            weights = np.array([
                        [[1/dn, 1/dn, 1/dn], [1/dn,  1/dn, 1/dn], [1/dn, 1/dn, 1/dn]],
                        [[1/dn, 1/dn, 1/dn], [1/dn, nm/dn, 1/dn], [1/dn, 1/dn, 1/dn]],
                        [[1/dn, 1/dn, 1/dn], [1/dn,  1/dn, 1/dn], [1/dn, 1/dn, 1/dn]],
                        ])
            ret = ndimage.convolve(ret, weights, mode='constant', cval=1.0)
        return ret

    def get_3d_magnitudes_from_curve(self, q, y):
        interp = interpolate.interp1d(q, y, kind='cubic', fill_value="extrapolate")
        qbin_values = interp(self.qbinsc)
        return self.get_3d_magnitudes(qbin_values)

    def get_3d_magnitudes_from_file(self, path):
        array, _ = np_loadtxt_robust(path)
        q = array[:,0]
        y = array[:,1]
        return self.get_3d_magnitudes_from_curve(q, y)



"""
    BoundedSecParams.py

    Copyright (c) 2022-2025, SAXS Team, KEK-PF
"""
import numpy as np
from .SimpleSecParams import initial_guess, sec_comformance
from molass_legacy.SecTheory.RetensionTime import estimate_conformance_params
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.SecTheory.MonoPore import MEMP_LIMIT

class BoundedSecParams:
    def __init__(self, *args):      # *args are not used
        self.init_method = initial_guess
        self.conf_method = self.bounded_sec_comformance
        self.estm_method = estimate_conformance_params
        self.nump_adjust = 0
        self.t0_upper_bound = get_setting("t0_upper_bound")

    def bounded_sec_comformance(self, xr_params, rg_params, seccol_params, poresize_bounds=None):
        conformance = sec_comformance(xr_params, rg_params, seccol_params, poresize_bounds=poresize_bounds)
        t0, K, rp, m = seccol_params
        penalty = max(0, t0 - self.t0_upper_bound)**2 + min(0, MEMP_LIMIT - m)**2 + min(0, m)**2
        return conformance + penalty



from scipy import optimize
from matplotlib import pylab as plt
import numpy as np
import pdb
from numpy import log

def exp_growth(t, x0, r):
    return x0 * ((1 + r) ** t)

def doubling_time(m, x_pts, y_pts):
    window = 10

    x1 = x_pts[m]
    y1 = y_pts[m]
    x2 = x_pts[m+window]
    y2 = y_pts[m+window]

    return (x2 - x1) * log(2) / log(y2 / y1)

# First, artificially create data points to work with
data_points = 42

# Create the x-axis
x_pts = range(0, data_points)

# Create noisy points with: y = x^2 + noise, with unique possible errors
y_pts = []
y_err = []
for i in range(data_points):
    random_scale = np.random.random()
    y_pts.append((i * i) + data_points * random_scale)
    y_err.append(random_scale * 100 + 100)

x_pts = np.array(x_pts)
y_pts = np.array(y_pts)
y_err = np.array(y_err)

# Fit to function
[x0, r], pcov  = optimize.curve_fit(exp_growth, x_pts, y_pts, p0=(0.001, 1.0))
fitted_data = exp_growth(x_pts, x0, r)

# Find doubling times
x_t2 = range(32)
t2 = []
t2_fit = []
for i in range(32):
    t2.append(doubling_time(i, x_pts, y_pts))
    t2_fit.append(doubling_time(i, x_pts, fitted_data))

# Plot
fig, (ax1, ax2, ax3) = plt.subplots(3, 1, sharex=True)

ax1.plot(x_pts, y_pts, 'bo')
ax1.errorbar(x_pts, y_pts, yerr=y_err)
ax1.set_ylim([0, 2000])
ax1.set_title('Artificially created raw data points with unique errors', fontsize=8)

ax2.plot(fitted_data, 'g-')
ax2.set_ylim([0, 2000])
ax2.set_title('Fitted exponential function', fontsize=8)

ax3.plot(x_t2, t2, 'ro', label='From points')
ax3.plot(x_t2, t2_fit, 'bo', label='From fitted')
ax3.set_title('Doubling time at each point (10 unit window)', fontsize=8)
ax3.legend(fontsize='8')

plt.show()
